"""Collection of general purpose functionalities."""

from .array_utils import *  # noqa: F401, F403
from .data_utils import *  # noqa: F401, F403
from .device_utils import *  # noqa: F401, F403
from .file_utils import *  # noqa: F401, F403
from .math_utils import *  # noqa: F401, F403
from .quantity_utils import *  # noqa: F401, F403
from .simulator_utils import *  # noqa: F401, F403



"""Modules necessary to run simulator.

Notes
-----
NECST modules should support simulator mode. Here is a collection of modules necessary
to implement the mode.

Important
---------
If there's a corresponding non-simulator script, functionalities in the original one
should be used as much as possible, through inheriting and overriding, to ensure the
behavior is the same as real one.

"""

from .antenna import *  # noqa: F401, F403



from abc import abstractmethod
from typing import Dict, List, Tuple

from ..device_base import DeviceBase


class Spectrometer(DeviceBase):
    @abstractmethod
    def get_spectra(self) -> Tuple[float, Dict[int, List[float]]]:
        """Timestamp and dict of spectral data for all boards."""
        ...



"""Custom Astropy units.

Notes
-----
Executing ``u.Unit("scan")`` without importing ``neclib`` will fail.
To use the units defined here, please import (contents of) ``neclib``, even if they
don't explicitly appear in your script.

"""

__all__ = ["scan", "point", "scan_to_point_equivalency"]

from typing import List

import astropy.units as u

from .typing import EquivalencyType


scan = u.def_unit("scan")
"""Custom unit to handle number of scan lines."""

point = u.def_unit("point")
"""Custom unit to handle number of pointing positions."""

dBm: u.DecibelUnit = u.dB(u.mW)
"""Custom unit to handle power in dBm."""

# Enable the use of the units in parsing it like ``u.Unit('scan')``, or other astropy
# functions e.g. ``scan.find_equivalent_units()``
u.add_enabled_units([scan, point])
u.add_enabled_aliases({"dBm": dBm})


def scan_to_point_equivalency(points_per_scan: int) -> List[EquivalencyType]:
    """Astropy units equivalency, between scan and point.

    Parameters
    ----------
    points_per_scan
        Number of observation points per 1 scan.

    Examples
    --------
    >>> eq = neclib.units.scan_to_point_equivalency(5)
    >>> u.Quantity("5scan").to(neclib.units.point, equivalencies=eq)
    25

    """
    return [(scan, point, lambda x: points_per_scan * x, lambda x: x / points_per_scan)]



from abc import abstractmethod
from typing import Literal

import astropy.units as u

from ..device_base import DeviceBase


class Motor(DeviceBase):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def set_step(self, step: int, axis: Literal["az", "el"]) -> None:
        """Drive to (maybe device-specific) absolute position."""
        ...

    @abstractmethod
    def set_speed(self, speed: float, axis: Literal["az", "el"]) -> None:
        ...

    @abstractmethod
    def get_step(self, axis: Literal["az", "el"]) -> int:
        """Maybe device-specific absolute position."""
        ...

    @abstractmethod
    def get_speed(self, axis: Literal["az", "el"]) -> u.Quantity:
        ...

    @abstractmethod
    def finalize(self) -> None:
        ...



r"""Additionally restrict the maximum drive speed near drive range limits.

If you understand the junior-high physics, you know the algorithm:

.. math::

   v &= \int a \ dt = a \int dt = at \qquad (v_0 = 0) \\
   x &= \int v \ dt = a \int t \ dt = \frac{a t^2}{2} \qquad (x_0 = 0) \\
   v(x) &= a \sqrt{\frac{2x}{a}} = \sqrt{2ax}

where :math:`x` is the distance between encoder reading and drive limits.

"""

from typing import Optional

import astropy.units as u

from ..typing import QuantityValue, UnitType
from .. import utils


class Decelerate:
    """Decelerate the telescope drive as it nears the drive range limits.

    Parameters
    ----------
    limit
        The range of encoder values that are considered the drive range limits.
    max_acceleration
        The maximum (absolute) acceleration of the telescope drive.

    Examples
    --------
    >>> limit = neclib.utils.ValueRange(5 << u.deg, 355 << u.deg)
    >>> calculator = neclib.safety.Decelerate(limit, 1.0 << u.deg / u.s**2)
    >>> calculator(354.6 << u.deg, 1 << u.deg / u.s)
    <Quantity 0.89442719 deg / s>
    >>> calculator(354.6 << u.deg, -1 << u.deg / u.s)
    <Quantity -1. deg / s>

    """

    def __init__(
        self,
        limit: utils.ValueRange[u.Quantity],  # type: ignore
        max_acceleration: u.Quantity,
    ) -> None:
        self.limit = limit
        self.max_acceleration = max_acceleration

    def __call__(
        self,
        encoder_reading: QuantityValue,
        velocity: QuantityValue,
        angle_unit: Optional[UnitType] = None,
    ) -> u.Quantity:
        velocity = utils.get_quantity(
            velocity, unit=(None if angle_unit is None else f"{angle_unit}/s")
        )
        encoder_reading = utils.get_quantity(encoder_reading, unit=angle_unit)

        if encoder_reading not in self.limit:
            return 0 << velocity.unit  # type: ignore

        position_relative_to_limits = (limit - encoder_reading for limit in self.limit)
        max_velocity_squared = (
            2 * self.max_acceleration * rel for rel in position_relative_to_limits
        )
        max_velocity = (
            v**0.5 if v.value == 0 else v / (abs(v) ** 0.5)  # type: ignore
            for v in max_velocity_squared
        )

        return utils.clip(velocity, *max_velocity)  # type: ignore



r"""PID controller for telescope main dish.

Optimum control parameter is calculated using a simple function consists of
Proportional, Integral and Derivative terms:

.. math::

    u(t) = K_\mathrm{p} \, e(t)
    + K_\mathrm{i} \int e(\tau) \, \mathrm{d}\tau
    + K_\mathrm{d} \frac{ \mathrm{d}e(t) }{ \mathrm{d}t }

where :math:`K_\mathrm{p}, K_\mathrm{i}` and :math:`K_\mathrm{d}` are non-negative
constants, :math:`u(t)` is the controller's objective parameter, and :math:`e(t)` is the
error between desired and actual values of explanatory parameter.

.. note::

    This script will be executed in high frequency with no vectorization, and there are
    many arrays updated frequently. These mean that the use of Numpy may not be the best
    choice to speed up the calculation. Measure the execution time first, then
    implement.

"""

__all__ = ["PIDController"]

import time
from contextlib import contextmanager
from typing import ClassVar, Dict, Generator, Literal, Tuple, Union

import astropy.units as u
import numpy as np

from .. import utils
from ..typing import AngleUnit
from ..utils import ParameterList

# Indices for parameter lists.
Last = -2
Now = -1


# Default values for PIDController.
DefaultK_p = 1.0
DefaultK_i = 0.5
DefaultK_d = 0.3
DefaultMaxSpeed = 2 << u.deg / u.s
DefaultMaxAcceleration = 2 << u.deg / u.s**2
DefaultErrorIntegCount = 50
DefaultThreshold = {
    "cmd_coord_change": 100 << u.arcsec,  # type: ignore
    "accel_limit_off": 20 << u.arcsec,  # type: ignore
    "target_accel_ignore": 2 << u.deg / u.s**2,
}
ThresholdKeys = Literal["cmd_coord_change", "accel_limit_off", "target_accel_ignore"]


class PIDController:
    """PID controller for telescope antenna.

    PID controller, a classical but sophisticated controller for system which has some
    delay on response to some input. This controller handles only 1 device, so use 2
    instances for Az-El control of telescope antenna.

    Parameters
    ----------
    pid_param
        Coefficients for PID formulation, list of [K_p, K_i, K_d].
    max_speed
        Maximum speed for telescope motion.
    max_acceleration
        Maximum acceleration for telescope motion.
    error_integ_count
        Number of error data to be stored for integral term calculation.
    threshold
        Thresholds for conditional executions. Interpreted keys are:

        - ``cmd_coord_change`` (angle)
            If separation between new command coordinate and last one is larger than
            this value, this controller assumes the target coordinate has been changed
            and resets the error integration.
        - ``accel_limit_off`` (angle)
            If separation between encoder reading and command coordinate is smaller than
            this value, this controller stops applying acceleration limit for quick
            convergence of drive.
        - ``target_accel_ignore`` (angular acceleration)
            If acceleration of target coordinate exceeds this value, the
            ``target_speed`` term is ignored as such commands most likely caused by
            software bug or network congestion.

    Attributes
    ----------

    k_p: float
        Proportional term coefficient.
    k_i: float
        Integral term coefficient.
    k_d: float
        Derivative term coefficient.
    k_c: float
        Constant term coefficient.
    max_speed: float
        Upper limit of drive speed in [``ANGLE_UNIT`` / s].
    max_acceleration: float
        Upper limit of drive acceleration in [``ANGLE_UNIT`` / s^2].
    error_integ_count: int
        Number of data stored for error integration.
    threshold: Dict[str, u.Quantity]
        Thresholds for conditional controls.
    cmd_speed: ParameterList[float]
        List of last 2 PID calculation results in unit [``ANGLE_UNIT`` / s].
    time: ParameterList[float]
        List of last ``error_integ_count`` UNIX timestamps the calculations are done.
    cmd_coord: ParameterList[float]
        List of last 2 command coordinates in [``ANGLE_UNIT``].
    enc_coord: ParameterList[float]
        List of last 2 encoder readings in [``ANGLE_UNIT``].
    error: ParameterList[float]
        List of last ``error_integ_count`` deviation values between ``cmd_coord`` and
        ``enc_coord``.
    target_speed: ParameterList[float]
        List of last 2 rate of change of command coordinates in [``ANGLE_UNIT`` / s].

    Notes
    -----
    This controller adds constant term to the general PID formulation. This is an
    attempt to follow constant motions such as raster scanning and sidereal motion
    tracking. ::

        speed = (K_c * target_speed)
            + (k_p * error)
            + (k_i * error_integral)
            + (k_d * error_derivative)

    This class keeps last 50 error values (by default) for integral term. This causes
    optimal PID parameters to change according to PID calculation frequency, as the time
    interval of the integration depends on the frequency.

    Tip
    ---
    All methods assume the argument is given in ``ANGLE_UNIT``, and return the results
    in that unit. If you need to change it, substitute ``ANGLE_UNIT`` before
    instantiating this class.

    Examples
    --------
    >>> neclib.controllers.PIDController.ANGLE_UNIT
    deg
    >>> controller = neclib.controllers.PIDController(
    ...     pid_param=[1.5, 0, 0],
    ...     max_speed="1000 arcsec/s",
    ...     max_acceleration="1.6 deg/s",
    ... )
    >>> target_coordinate, encoder_reading = 30, 10  # deg
    >>> controller.get_speed(target_coordinate, encoder_reading)
    1.430511474609375e-05  # deg/s
    >>> controller.get_speed(target_coordinate, encoder_reading)
    0.20356178283691406  # deg/s

    """

    ANGLE_UNIT: ClassVar[AngleUnit] = "deg"
    """Unit in which all public functions accept as its argument and return."""

    def __init__(
        self,
        *,
        pid_param: Tuple[float, float, float] = (DefaultK_p, DefaultK_i, DefaultK_d),
        max_speed: Union[str, u.Quantity] = DefaultMaxSpeed,
        max_acceleration: Union[str, u.Quantity] = DefaultMaxAcceleration,
        error_integ_count: int = DefaultErrorIntegCount,
        threshold: Dict[ThresholdKeys, Union[str, u.Quantity]] = DefaultThreshold,  # type: ignore  # noqa: E501
    ) -> None:
        self.k_p, self.k_i, self.k_d = pid_param
        self.k_c = 1
        self.max_speed = utils.parse_quantity(max_speed, unit=self.ANGLE_UNIT).value
        self.max_acceleration = utils.parse_quantity(
            max_acceleration, unit=self.ANGLE_UNIT
        ).value
        self.error_integ_count = error_integ_count
        _threshold = DefaultThreshold.copy()
        _threshold.update(threshold)
        self.threshold = {
            k: utils.parse_quantity(v, unit=self.ANGLE_UNIT).value
            for k, v in _threshold.items()
        }

        # Initialize parameter buffers.
        self._initialize()

    @property
    def dt(self) -> float:
        """Time interval of last 2 PID calculations."""
        return self.time[Now] - self.time[Last]

    @property
    def error_integral(self) -> float:
        """Integral of error."""
        _time, _error = np.array(self.time), np.array(self.error)
        dt = _time[1:] - _time[:-1]
        error_interpolated = (_error[1:] + _error[:-1]) / 2
        return np.nansum(error_interpolated * dt)  # type: ignore

    @property
    def error_derivative(self) -> float:
        """Derivative of error."""
        return (self.error[Now] - self.error[Last]) / self.dt

    def _set_initial_parameters(self, cmd_coord: float, enc_coord: float) -> None:
        """Initialize parameters, except necessity for continuous control."""
        self._initialize()

        if np.isnan(self.cmd_speed[Now]):
            self.cmd_speed.push(0)
        self.time.push(time.time())
        self.cmd_coord.push(cmd_coord)
        self.enc_coord.push(enc_coord)
        self.error.push(cmd_coord - enc_coord)
        self.target_speed.push(0)

    def _initialize(self) -> None:
        """Define control loop parameters."""
        if not hasattr(self, "cmd_speed"):
            self.cmd_speed = ParameterList.new(2)
        self.time = ParameterList.new(2 * int(self.error_integ_count / 2))
        self.cmd_coord = ParameterList.new(2)
        self.enc_coord = ParameterList.new(2)
        self.error = ParameterList.new(2 * int(self.error_integ_count / 2))
        self.target_speed = ParameterList.new(2)

    def get_speed(
        self,
        cmd_coord: float,
        enc_coord: float,
        stop: bool = False,
    ) -> float:
        """Modulated drive speed.

        Parameters
        ----------
        cmd_coord
            Instructed Az/El coordinate.
        enc_coord
            Az/El encoder reading.
        stop
            If ``True``, the telescope won't move regardless of the inputs.

        """
        delta_cmd_coord = cmd_coord - self.cmd_coord[Now]
        if np.isnan(self.time[Now]) or (
            abs(delta_cmd_coord) > self.threshold["cmd_coord_change"]
        ):
            self._set_initial_parameters(cmd_coord, enc_coord)
            # Set default values on initial run or on detection of sudden jump of error,
            # which may indicate a change of command coordinate.
            # This will give too small `self.dt` later, but that won't propose any
            # problem, since `current_speed` goes to 0, and too large target_speed will
            # be ignored in `_calc_pid`.

        current_speed = self.cmd_speed[Now]
        # Encoder readings cannot be used, due to the lack of stability.

        self.time.push(time.time())
        self.cmd_coord.push(cmd_coord)
        self.enc_coord.push(enc_coord)
        self.error.push(cmd_coord - enc_coord)
        self.target_speed.push((self.cmd_coord[Now] - self.cmd_coord[Last]) / self.dt)

        # Calculate and validate drive speed.
        speed = self._calc_pid()
        if abs(self.error[Now]) > self.threshold["accel_limit_off"]:
            # When error is small, smooth control delays the convergence of drive.
            # When error is large, smooth control can avoid overshooting.
            max_diff = self.max_acceleration * self.dt
            # 0.2 clipping is to avoid large acceleration caused by large dt.
            speed = utils.clip(
                speed,  # type: ignore
                current_speed - max_diff,
                current_speed + max_diff,
            )  # Limit acceleration.
        # Limit speed.
        speed = utils.clip(speed, absmax=self.max_speed)  # type: ignore

        if stop:
            self.cmd_speed.push(0)
        else:
            self.cmd_speed.push(speed)

        return self.cmd_speed[Now]

    def _calc_pid(self) -> float:
        # Speed of the move of commanded coordinate. This includes sidereal motion, scan
        # speed, and other non-static component of commanded value.
        target_acceleration = (
            self.target_speed[Now] - self.target_speed[Last]
        ) / self.dt
        if abs(target_acceleration) > self.threshold["target_accel_ignore"]:
            self.target_speed[Now] = 0

        return (
            self.k_c * self.target_speed[Now]
            + self.k_p * self.error[Now]
            + self.k_i * self.error_integral
            + self.k_d * self.error_derivative
        )

    @contextmanager
    def params(self, **kwargs) -> Generator[None, None, None]:
        original_attrs = {
            "k_p": self.k_p,
            "k_i": self.k_i,
            "k_d": self.k_d,
            "k_c": self.k_c,
            "max_speed": self.max_speed,
            "max_acceleration": self.max_acceleration,
            "error_integ_count": self.error_integ_count,
        }
        original_thresholds = self.threshold.copy()

        for k, v in kwargs.items():
            if k in original_attrs:
                setattr(self, k, v)
            elif k in original_thresholds:
                self.threshold[k] = v
            else:
                raise ValueError(f"Invalid parameter: {k}")

        try:
            yield
        finally:
            for k in original_attrs.keys():
                setattr(self, k, original_attrs[k])
            for k in original_thresholds.keys():
                self.threshold[k] = original_thresholds[k]



__all__ = ["TR73U"]

import struct
from typing import Dict

import astropy.units as u
import ogameasure

from ... import get_logger, utils
from .weather_station_base import WeatherStation


class TR73U(WeatherStation):

    Manufacturer = "TandD"
    Model = "TR73U"

    Identifier = "port"

    def __init__(self) -> None:
        self.logger = get_logger(__name__)
        self.ondotori = ogameasure.TandD.tr_73u(self.Config.port)

    def _get_data(self) -> Dict[str, float]:
        with utils.busy(self, "busy"):
            try:
                return self.ondotori.output_current_data()
            except struct.error:
                self.logger.warning("Failed to get data from TR73U")
                return {"temp": 0.0, "humid": 0.0, "press": 0.0}

    def get_temp(self) -> u.Quantity:
        data = self._get_data()
        data_K = (data["temp"] * u.deg_C).to(u.K, equivalencies=u.temperature())
        return data_K

    def get_humid(self) -> float:
        data = self._get_data()
        return data["humid"] * 0.01

    def get_press(self) -> u.Quantity:
        data = self._get_data()
        return data["press"] * u.hPa

    def finalize(self) -> None:
        self.ondotori.ser.close()



import astropy.units as u
import ogameasure

from .vacuum_gauge_base import VacuumGauge


class TPG261(VacuumGauge):

    Model = "TPG261"
    Manufacturer = "Pfeiffer"

    Identifier = "host"

    def __init__(self):
        com = ogameasure.ethernet(self.Config.host, self.Config.port)
        self.io = ogameasure.Pfeiffer.tpg261_lan(com)
        self.io.pres_unit_pa()

    def get_pressure(self) -> u.Quantity:
        return self.io.pressure() * u.Pa

    def finalize(self):
        self.io.com.close()



"""Utility functions for real-time control logics."""

__all__ = ["busy"]

import time


class busy:
    """Manages the busy state of the object.

    This context manager blocks and stops the execution of its contents until the object
    confirmed to be not busy. When the context manager exits, it frees the busy flag.
    This functionality would be useful to avoid conflicting operations to be executed at
    the same time.

    Parameters
    ----------
    obj
        The object to manage the busy state of.
    flagname
        The name of the attribute to use to store the busy state.

    Examples
    --------
    >>> class Foo:
    ...     def task_a(self):
    ...         with busy(self, "busy"):
    ...             print("Task A")
    ...             time.sleep(1)
    ...             print("Task A done")
    ...     def task_b(self):
    ...         with busy(self, "busy"):
    ...             print("Task B")
    ...             time.sleep(0.5)
    ...             print("Task B done")
    >>> with concurrent.futures.ThreadPoolExecutor() as executor:
    ...     foo = Foo()
    ...     future1 = executor.submit(foo.task_a)
    ...     future2 = executor.submit(foo.task_b)
    ...     concurrent.futures.wait([future1, future2])

    The above example attempts to execute two tasks concurrently. However, the ``busy``
    context prevents the second task from starting until the first task finishes.

    Attention
    ---------
    The ``flagname`` argument must be the same across all ``busy`` contexts which blocks
    each other. To manage multiple busy states, use different ``flagname``-s to
    distinguish the task groups.

    """

    class __busyflag:
        def __init__(self, busy: bool = False):
            self.busy = busy

    def __init__(self, obj: object, flagname: str) -> None:
        if hasattr(obj, flagname) and (
            not isinstance(getattr(obj, flagname), self.__busyflag)
        ):
            raise ValueError(f"Object {obj!r} already has attribute {flagname!r}")
        self.__obj = obj
        self.__flagname = flagname

    @property
    def busy(self) -> bool:
        if not hasattr(self.__obj, self.__flagname):
            object.__setattr__(self.__obj, self.__flagname, self.__busyflag())
        return getattr(self.__obj, self.__flagname).busy

    @busy.setter
    def busy(self, value: bool) -> None:
        getattr(self.__obj, self.__flagname).busy = value

    def __enter__(self) -> None:
        while self.busy:
            time.sleep(0.05)
        self.busy = True

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.busy = False
        delattr(self.__obj, self.__flagname)
        if exc_type is not None:
            raise



from abc import abstractmethod
from typing import Literal

import astropy.units as u

from ..device_base import DeviceBase


class Encoder(DeviceBase):

    # Manufacturer: str = ""
    # Model: str

    @abstractmethod
    def get_reading(self, axis: Literal["az", "el"]) -> u.Quantity:
        ...

    @abstractmethod
    def finalize(self) -> None:
        ...



"""Emulator for telescope antenna motion and corresponding encoder reading."""

__all__ = ["AntennaEncoderEmulator"]

import time
from typing import Callable, ClassVar, List, Literal, Tuple, Union

import astropy.units as u
import numpy as np

from .. import utils
from ..typing import AngleUnit
from ..utils import AzElData, ParameterList

# Indices for parameter lists.
Last = -2
Now = -1

QtyFn = Union[u.Quantity, Callable[[u.Quantity, u.Quantity], u.Quantity]]

DefaultMomentOfInertia: Tuple[QtyFn, QtyFn] = (
    lambda az, el: (2500 + 3000 * np.cos(el)) * u.Unit("kg m2"),
    3000 * u.Unit("kg m2"),
)  # NANTEN2, rough estimation based on Ito (2005), Master's thesis.
DefaultMotorTorque = (11.5 * u.Unit("N m"), 11.5 * u.Unit("N m"))
# NANTEN2, from Ito (2005), Master's thesis.
DefaultAngularResolution = (
    360 * 3600 / (23600 * 400) * u.arcsec,  # type: ignore
    360 * 3600 / (23600 * 400) * u.arcsec,  # type: ignore
)  # RENISHAW RESM series encoder, which is installed in NANTEN2.

InitialPosition = AzElData(180 << u.deg, 45 << u.deg)  # type: ignore


class AntennaEncoderEmulator:
    """Emulator of antenna response for simulator.

    Parameters
    ----------
    device_moment_of_inertia
        Moment of inertia of antenna.
    motor_torque
        Maximum torque the motor can exert.
    angular_resolution
        Angular resolution of the encoder.

    Notes
    -----
    To support coordinate-dependent moment of inertia, Az. and El. axes cannot
    independently be implemented.

    .. warning::

        API may change, as API of encoder driver module isn't fixed yet.

    Example
    -------
    >>> enc = neclib.simulators.AntennaEncoderEmulator()
    >>> pid_az = neclib.controllers.PIDController()
    >>> speed = pid_az.get_speed(30, enc.read.az)
    >>> enc.command(speed, "az")

    """

    ANGLE_UNIT: ClassVar[AngleUnit] = "deg"

    def __init__(
        self,
        device_moment_of_inertia: Tuple[QtyFn, QtyFn] = DefaultMomentOfInertia,
        motor_torque: Tuple[u.Quantity, u.Quantity] = DefaultMotorTorque,
        angular_resolution: Tuple[u.Quantity, u.Quantity] = DefaultAngularResolution,
    ) -> None:
        device_moment_of_inertia = self._make_callable(
            device_moment_of_inertia  # type: ignore
        )

        # Force the use of SI unit
        device_moment_of_inertia = tuple(
            lambda az, el: func(az, el).si.value  # type: ignore
            for func in device_moment_of_inertia
        )
        motor_torque = tuple(qty.si.value for qty in motor_torque)
        angular_resolution = tuple(
            qty.to(self.ANGLE_UNIT).value for qty in angular_resolution
        )

        self.moment_of_inertia = AzElData(*device_moment_of_inertia)
        self.torque = AzElData(*motor_torque)
        self.angular_resolution = AzElData(*angular_resolution)

        self.time = ParameterList.new(2, time.time())
        self.position = AzElData(
            InitialPosition.az.to(self.ANGLE_UNIT).value,
            InitialPosition.el.to(self.ANGLE_UNIT).value,
        )
        self.speed = AzElData(0, 0)
        self.cmd_speed = AzElData(0, 0)

        # Conversion factor for fluctuation and angular acceleration scaling.
        self.rad = utils.angle_conversion_factor("rad", self.ANGLE_UNIT)
        self.arcsec = utils.angle_conversion_factor("arcsec", self.ANGLE_UNIT)

    @staticmethod
    def _make_callable(
        value_or_function: List[Union[u.Quantity, QtyFn]]
    ) -> List[QtyFn]:
        """Make common interface for value acquisition."""
        return [
            item if callable(item) else lambda az, el: item
            for item in value_or_function
        ]  # type: ignore

    @property
    def abs_acceleration(self) -> AzElData:
        """Absolute value of angular acceleration at the moment."""
        try:
            return AzElData(
                self.torque.az
                / self.moment_of_inertia.az(self.position.az, self.position.el)
                * self.rad,
                self.torque.el
                / self.moment_of_inertia.el(self.position.az, self.position.el)
                * self.rad,
            )
        except ZeroDivisionError:
            small_mom_of_inertia = 1e-8
            return AzElData(
                self.torque.az / small_mom_of_inertia * self.rad,
                self.torque.el / small_mom_of_inertia * self.rad,
            )

    @property
    def dt(self) -> float:
        """Time interval between last two call for encoder reading."""
        return self.time[Now] - self.time[Last]

    def read(self) -> AzElData:
        """Get current encoder reading.

        Notes
        -----
        Acceleration during consecutive calls are approximated to be constant.

        Examples
        --------
        >>> v = enc.read()
        >>> v.az
        12.3

        """
        abs_accel = self.abs_acceleration
        now = time.time()
        self.time.push(now)

        accel = AzElData(
            np.sign(self.cmd_speed.az - self.speed.az) * abs_accel.az,
            np.sign(self.cmd_speed.el - self.speed.el) * abs_accel.el,
        )
        _speed = AzElData(
            self.speed.az + accel.az * self.dt,
            self.speed.el + accel.el * self.dt,
        )  # No consideration on the behavior when current speed reached the command.
        sped_over = AzElData(
            max(0, abs(_speed.az) - abs(self.cmd_speed.az)),
            max(0, abs(_speed.el) - abs(self.cmd_speed.el)),
        )  # How much over-sped when constant acceleration is assumed.
        accel0_duration = AzElData(
            sped_over.az / abs_accel.az, sped_over.el / abs_accel.el
        )  # How long the acceleration should be set to 0 to sustain command speed.
        next_position = AzElData(
            accel.az * self.dt**2 / 2
            + self.speed.az * self.dt
            + self.position.az
            - sped_over.az * accel0_duration.az / 2,
            accel.el * self.dt**2 / 2
            + self.speed.el * self.dt
            + self.position.el
            - sped_over.el * accel0_duration.el / 2,
        )

        self.speed.az = utils.clip(_speed.az, absmax=self.cmd_speed.az)
        self.speed.el = utils.clip(_speed.el, absmax=self.cmd_speed.el)
        self.position.az = next_position.az
        self.position.el = next_position.el

        fluctuation = np.random.randn
        return AzElData(
            utils.discretize(
                0.1 * fluctuation() * self.arcsec + self.position.az,
                step=self.angular_resolution.az,
            ),
            utils.discretize(
                0.1 * fluctuation() * self.arcsec + self.position.el,
                step=self.angular_resolution.el,
            ),
        )

    def command(self, speed: float, axis: Literal["az", "el"]) -> None:
        """Set angular speed of intention with angular unit ``ANGLE_UNIT``.

        Parameters
        ----------
        speed
            Angular speed, which may be calculated by PID controller.
        axis
            Controlling altazimuth axis.

        """
        setattr(self.cmd_speed, axis, speed)



from collections import defaultdict
from functools import partial
from types import SimpleNamespace
from typing import Dict, List


class Status:
    """Status manager.

    Parameters
    ----------
    levels
        List of level names, default is ["warning", "critical"]. The names should be
        given in ascending order in severity.

    Raises
    ------
    ValueError
        If (at least) one of level names are reserved by Python.

    Examples
    --------
    >>> status = Status()
    >>> status["topic1"].warning = True
    >>> status["topic1"]
    namespace(warning=True, critical=False)
    >>> status["topic2"] = {"warning": False, "critical": False}
    >>> status.warning()
    True
    >>> status.critical()
    False

    """

    def __init__(self, levels: List[str] = ["warning", "critical"]) -> None:
        # Check if levels are assignable
        reserved_attr = [
            level for level in levels if hasattr(self, level) or level.startswith("__")
        ]
        if len(reserved_attr) > 0:
            raise ValueError(f"Level names {reserved_attr} not assignable")

        self.__status: Dict[str, SimpleNamespace] = defaultdict(
            lambda: SimpleNamespace(**{k: None for k in levels})
        )
        self.__levels = levels

        # Assign property-like checkers
        for level in levels:
            setattr(self, level, partial(self.__get_status, level))

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(severity: {' < '.join(self.__levels)})"

    def __get_status(self, level: str) -> bool:
        """Check if any topic has set to the level."""
        if any([getattr(v, level) for v in self.__status.values()]):
            return True  # If any topic is in this exact level, return True

        next_level_idx = self.__levels.index(level) + 1
        if next_level_idx == len(self.__levels):
            return False  # No truthy value found even in severer levels
        return self.__get_status(self.__levels[next_level_idx])  # Check severer levels

    def __getitem__(self, key):
        return self.__status[key]

    def __setitem__(self, key, value):
        if not isinstance(value, dict):
            raise TypeError("Value must be dict")
        value = {k: v for k, v in value.items() if k in self.__status[key].__dict__}
        self.__status[key].__dict__.update(value)



__all__ = ["config", "configure"]

import os
import re
import shutil
from collections.abc import ItemsView, KeysView, ValuesView
from pathlib import Path
from typing import Any, Callable, Generic, List, Optional, Tuple, Type, TypeVar, Union

import astropy.units as u
from astropy.coordinates import EarthLocation
from tomlkit.toml_file import TOMLFile

from . import EnvVarName, get_logger
from .exceptions import NECSTConfigurationError
from .utils import ValueRange, read_file

logger = get_logger(__name__)

DefaultNECSTRoot = Path.home() / ".necst"
DefaultConfigPath = DefaultNECSTRoot / "config.toml"

T = TypeVar("T")
V = TypeVar("V")


class Parameter(Generic[T, V]):
    def __init__(self, parser: Callable[[T], V]) -> None:
        self._parser = parser

    def __set_name__(self, owner: Type[object], name: str) -> None:
        self._name = name

    @property
    def _private_name(self):
        return "_" + self._name

    def __get__(
        self, obj: object, objtype: Optional[Type[object]] = None
    ) -> Optional[V]:
        if obj is None:
            return None
        return getattr(obj, self._private_name)

    def __set__(self, obj: object, value: T) -> None:
        if type(getattr(obj, self._name, None)) not in [type(None), self.__class__]:
            raise AttributeError(f"Cannot overwrite existing attribute {self._name}")
        parsed = self._parser(value)
        if isinstance(parsed, Path):  # Handle relative path, inside dedicated directory
            parsed = getattr(obj, "_dotnecst", DefaultNECSTRoot) / parsed
        setattr(obj, self._private_name, parsed)


class Configuration:
    """NECST configuration.

    This parameter collection supports flexible name look-up. If you query the parameter
    'antenna' by ``config.antenna``, all parameters prefixed by 'antenna' will be
    extracted.

    Examples
    --------
    >>> neclib.config.observatory
    'OMU1P85M'
    >>> neclib.config.antenna_pid_param_az
    [1.5, 0.0, 0.0]
    >>> neclib.config.antenna_pid_param
    SimpleNamespace(az=[1.5, 0.0, 0.0], el=[1.5, 0.0, 0.0])
    >>> neclib.config.antenna_pid_param.az
    [1.5, 0.0, 0.0]

    """

    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        self._config = None
        self.reload()

    def reload(self) -> None:
        """Reload the parameters, in case the config file is updated.

        Warning
        -------
        Parameters set manually (like `config.<parameter_name> = ...`) will be lost on
        reload.

        """
        self.__config_path = self.__find_config_path()
        params = TOMLFile(self.__config_path).read().unwrap()
        new_config = _Cfg(self, **params)
        # Attribute validation
        attr_cross_section = set(dir(self.__class__)) & set(dir(new_config))
        dupl_attrs = list(filter(lambda x: not x.startswith("_"), attr_cross_section))
        if dupl_attrs:
            raise NECSTConfigurationError(
                f"Parameters {dupl_attrs} cannot be set; reserved name."
            )
        # Assign after validation
        self.__config = new_config

    @property
    def _dotnecst(self) -> Path:
        return self.__config_path.parent

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(file='{self.__config_path}')"

    def __str__(self) -> str:
        return str(self.__config)

    def __getattr__(self, key: str) -> Optional[Any]:
        return getattr(self.__config, key, None)

    def __find_config_path(self) -> Path:
        root_candidates = [DefaultNECSTRoot]
        if EnvVarName.necst_root in os.environ:
            root_candidates.insert(0, os.environ[EnvVarName.necst_root])
        config_file_name = os.environ.get(EnvVarName.necst_config_name, "config.toml")
        candidates = map(lambda x: str(x) + f"/{config_file_name}", root_candidates)

        for path in candidates:
            try:
                read_file(path, localonly=True)
                found = Path(re.sub(r".*://", "", path))
                logger.info(f"Imported configuration file '{found}'")
                return found
            except FileNotFoundError:
                pass

            try:
                saveto = DefaultNECSTRoot / "config.toml"
                read_file(path, localonly=False, saveto=saveto, overwrite=True)
                logger.info(
                    f"Imported configuration file '{path}', "
                    f"which was saved to '{saveto}'"
                )
                return saveto
            except FileNotFoundError:
                pass

        logger.error(
            "Config file not found, using the default parameters. "
            "To create the file with default parameters, run `neclib.configure()`."
        )
        return Path(__file__).parent / "src" / "config.toml"

    @classmethod
    def configure(cls) -> None:
        """Create config file under ``$HOME/.necst``"""
        DefaultNECSTRoot.mkdir(exist_ok=True)
        for filename in ["config.toml", "pointing_param.toml"]:
            _target_path = DefaultNECSTRoot / filename
            if _target_path.exists():
                logger.error(f"'{_target_path}' already exists, skipping...")
                continue
            shutil.copyfile(Path(__file__).parent / "src" / filename, _target_path)
        cls().reload()


class _Cfg:

    observatory = Parameter(str)
    location = Parameter(lambda x: EarthLocation(**x))
    observation_frequency = Parameter(u.Quantity)
    simulator = Parameter(bool)
    record_root = Parameter(Path)
    alert_interval_sec = Parameter(float)
    antenna_pid_param_az = Parameter(list)
    antenna_pid_param_el = Parameter(list)
    antenna_drive_range_az = Parameter(lambda x: ValueRange(*map(u.Quantity, x)))  # type: ignore  # noqa: E501
    antenna_drive_range_el = Parameter(lambda x: ValueRange(*map(u.Quantity, x)))  # type: ignore  # noqa: E501
    antenna_drive_warning_limit_az = Parameter(lambda x: ValueRange(*map(u.Quantity, x)))  # type: ignore  # noqa: E501
    antenna_drive_warning_limit_el = Parameter(lambda x: ValueRange(*map(u.Quantity, x)))  # type: ignore  # noqa: E501
    antenna_drive_critical_limit_az = Parameter(lambda x: ValueRange(*map(u.Quantity, x)))  # type: ignore  # noqa: E501
    antenna_drive_critical_limit_el = Parameter(lambda x: ValueRange(*map(u.Quantity, x)))  # type: ignore  # noqa: E501
    antenna_pointing_accuracy = Parameter(u.Quantity)
    antenna_pointing_parameter_path = Parameter(Path)
    antenna_scan_margin = Parameter(u.Quantity)
    antenna_max_acceleration_az = Parameter(u.Quantity)
    antenna_max_acceleration_el = Parameter(u.Quantity)
    antenna_max_speed_az = Parameter(u.Quantity)
    antenna_max_speed_el = Parameter(u.Quantity)
    antenna_speed_to_pulse_factor_az = Parameter(u.Quantity)
    antenna_speed_to_pulse_factor_el = Parameter(u.Quantity)
    antenna_command_frequency = Parameter(int)
    antenna_command_offset_sec = Parameter(float)
    ros_service_timeout_sec = Parameter(float)
    ros_communication_deadline_sec = Parameter(float)
    ros_logging_interval_sec = Parameter(float)
    ros_topic_scan_interval_sec = Parameter(float)

    def __init__(
        self, config_manager: Configuration, prefix: str = "", /, **kwargs
    ) -> None:
        keys = set(k.lower() for k in kwargs.keys())
        if len(keys) != len(kwargs):
            dupl = set(kwargs) - keys
            raise NECSTConfigurationError(
                f"Parameters {dupl} cannot be assigned; duplicate definition."
            )

        self.__config_manager = config_manager
        self.__prefix = prefix
        self.__parameters = [
            self.__assign_parameter(key, value) for key, value in kwargs.items()
        ]

    def keys(self, full: bool = False) -> KeysView:
        prefix_length = len(self.__prefix)
        if full:
            p = self.__parameters
        else:
            p = map(lambda x: (x[0][prefix_length:], x[1]), self.__parameters)
        return dict(p).keys()

    def values(self) -> ValuesView:
        prefix_length = len(self.__prefix)
        p = map(lambda x: (x[0][prefix_length:], x[1]), self.__parameters)
        return dict(p).values()

    def items(self, full: bool = False) -> ItemsView:
        prefix_length = len(self.__prefix)
        if full:
            p = self.__parameters
        else:
            p = map(lambda x: (x[0][prefix_length:], x[1]), self.__parameters)
        return dict(p).items()

    @property
    def _dotnecst(self) -> Path:
        return self.__config_manager._dotnecst

    @property
    def _prefix(self) -> str:
        return self.__prefix

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(prefix='{self.__prefix}')"

    def __str__(self) -> str:
        width = max(len(p) for p, _ in self.__parameters) - len(self.__prefix) + 2

        def _prettify(key: str) -> str:
            key = key[len(self.__prefix) :]
            value = getattr(self, key, None)
            return f"    {key:{width}s}{value!s}    ({type(value).__name__})"

        _parameters = "\n".join([_prettify(k) for k, _ in self.__parameters])
        return f"NECST configuration (prefix='{self.__prefix}')\n{_parameters}"

    def __assign_parameter(self, k, v) -> Tuple[str, Any]:
        if k.startswith("_"):
            raise NECSTConfigurationError(
                f"Parameter '{k!r}' cannot be assigned; name starts with '_' is invalid"
            )
        k = self.__prefix + k
        if k.lower() in self.__reserved_names:
            raise NECSTConfigurationError(
                f"Parameter {k!r} cannot be assigned; reserved name."
            )
        setattr(self, k, v)
        return (k, getattr(self, k))

    @property
    def __reserved_names(self) -> List[str]:
        return [
            k.lower() for k, v in vars(self).items() if not isinstance(v, Parameter)
        ]

    def __getattr__(self, key: str) -> Optional[Union[Any, "_Cfg"]]:
        if key.startswith("_"):
            raise AttributeError(f"Attribute '{key}' not found.")

        value = None
        if not key.islower():
            value = getattr(self, key.lower(), None)
        if (value is None) and self.__prefix and (not key.startswith(self.__prefix)):
            value = getattr(self, self.__prefix + key, None)
        if value is None:
            prefix = key + "_"
            prefix_length = len(prefix)
            _match = {
                k[prefix_length:]: v
                for k, v in self.__parameters
                if k.lower().startswith(prefix.lower())
            }
            value = _Cfg(self.__config_manager, prefix, **_match) if _match else None

        return value

    def __gt__(self, other: Any) -> bool:
        if other is None:
            return True

        if not isinstance(other, _Cfg):
            return NotImplemented

        if set(self.keys()) <= set(other.keys()):
            return False

        lazy_eq = (getattr(self, k) == getattr(other, k) for k in other.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in other.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __lt__(self, other: Any) -> bool:
        if other is None:
            return False

        if not isinstance(other, _Cfg):
            return NotImplemented

        if set(self.keys()) >= set(other.keys()):
            return False

        lazy_eq = (getattr(self, k) == getattr(other, k) for k in self.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in self.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __eq__(self, other: Any) -> bool:
        if other is None:
            return False

        if not isinstance(other, _Cfg):
            return NotImplemented

        if set(self.keys()) != set(other.keys()):
            return False

        lazy_eq = (getattr(self, k) == getattr(other, k) for k in self.keys())
        full_eq = (getattr(self, k) == getattr(other, k) for k in self.keys(full=True))
        return True if all(lazy_eq) or all(full_eq) else False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __ge__(self, other: Any) -> bool:
        return self.__eq__(other) or self.__gt__(other)

    def __le__(self, other: Any) -> bool:
        return self.__eq__(other) or self.__lt__(other)


config = Configuration()
configure = Configuration.configure



"""Controllers for miscellaneous devices.

Controller may refer to collection of functions to control devices or implementation of
some algorithms to control something.

"""

from .antenna_pid import *  # noqa: F401, F403



__all__ = ["linear_sequence"]

from typing import TypeVar, Union, overload

import astropy.units as u
import numpy as np

from ..typing import Number

T = TypeVar("T", Number, u.Quantity)


@overload
def linear_sequence(
    start: Union[int, float], stop: Union[int, float], num: int
) -> np.ndarray:
    ...


@overload
def linear_sequence(start: u.Quantity, stop: u.Quantity, num: int) -> u.Quantity:
    ...


def linear_sequence(start: T, step: T, num: int) -> T:
    """Similar to `numpy.linspace`, but unknown terminal value."""
    start, step = np.asanyarray(start), np.asanyarray(step)
    return np.linspace(start, start + step * (num - 1), num)



"""Utility functions for physical quantity or unit handling."""

__all__ = [
    "angle_conversion_factor",
    "parse_quantity",
    "partially_convert_unit",
    "force_data_type",
    "quantity2builtin",
    "dAz2dx",
    "dx2dAz",
    "get_quantity",
]

import math
from typing import Any, Dict, Hashable, Optional, List, Union, overload

import astropy.units as u

from ..typing import AngleUnit, Number, UnitType


def angle_conversion_factor(original: AngleUnit, to: AngleUnit) -> float:
    """Conversion factor between angular units.

    Parameters
    ----------
    original
        Original angular unit.
    to
        Unit to convert to.

    Notes
    -----
    More general implementation may be realized using astropy.units, but it's ~1000
    times slower than this thus can be a bottleneck in time critical calculations.

    Examples
    --------
    >>> angle_deg = 1
    >>> angle_deg * neclib.utils.angle_conversion_factor("deg", "arcsec")
    3600  # arcsec

    """
    equivalents = {"deg": 1, "arcmin": 60, "arcsec": 3600, "rad": math.pi / 180}
    try:
        return equivalents[to] / equivalents[original]
    except KeyError:
        raise ValueError(
            "Units other than than 'deg', 'arcmin' or 'arcsec' are not supported."
        )


def force_data_type(az, el, unit="deg"):
    def convert_unit(param):
        if isinstance(param, u.Quantity):
            param.to(unit)
        else:
            param *= u.Unit(unit)
        return param

    az = convert_unit(az)
    el = convert_unit(el)

    return az, el


def parse_quantity(
    quantity: Union[str, u.Quantity], *, unit: Optional[UnitType] = None
) -> u.Quantity:
    """Get ``astropy.units.Quantity`` object, optionally converting units.

    Parameters
    ----------
    quantity
        ``Quantity`` object or its ``str`` expression.
    unit
        Unit(s) you want to use.

    Examples
    --------
    >>> neclib.utils.parse_quantity("3 M_sun pc^-2")
    <Quantity 3. solMass / pc2>
    >>> neclib.utils.parse_quantity("3 M_sun pc^-2", unit="kg")
    <Quantity 5.96542625e+30 kg / pc2>

    See Also
    --------
    partially_convert_unit : For unit conversion.

    """
    if unit is None:
        return u.Quantity(quantity)
    else:
        return partially_convert_unit(u.Quantity(quantity), unit)


def partially_convert_unit(quantity: u.Quantity, new_unit: UnitType) -> u.Quantity:
    """Replace unit of given dimension.

    Parameters
    ----------
    quantity
        Original ``Quantity`` object.
    new_unit
        Unit(s) to employ.

    Notes
    -----
    ``new_unit`` should be (product of) units which construct ``quantity``'s unit. If
    unit of ``quantity`` is ``J / s``, this function can employ ``erg``, but cannot
    convert to ``W``.

    Examples
    --------
    >>> quantity = u.Quantity("3 L_sun s")
    >>> neclib.utils.partially_convert_unit(quantity, "W")
    <Quantity 1.1484e+27 s W>
    >>> neclib.utils.partially_convert_unit(quantity, "W hour")
    <Quantity 3.19e+23 h W>
    >>> neclib.utils.partially_convert_unit(quantity, "J")
    ValueError: Couldn't find equivalent units; give equivalent(s) of ["s", "solLum"].

    """
    base_units = quantity.unit.bases  # type: ignore
    new_units: List[u.UnitBase] = u.Unit(new_unit).bases  # type: ignore
    for base in base_units:
        if not any([base.is_equivalent(new_u) for new_u in new_units]):
            new_units.append(base)
    if len(base_units) != len(new_units):
        raise ValueError(
            f"Couldn't find equivalent units; give equivalent(s) of {base_units}."
        )
    return quantity.decompose(bases=new_units)


def quantity2builtin(
    quantity: Dict[Hashable, u.Quantity],
    unit: Union[Dict[Hashable, Union[u.Unit, str]], Dict[UnitType, List[str]]] = {},
) -> Dict[Hashable, Union[int, float, Any]]:
    """Convert quantity to Python's built-in types.

    Parameters
    ----------
    quantity
        Dictionary of quantity objects to convert.
    unit
        Dictionary of units to employ, in either formats: ``{parameter_name: unit}``,
        ``{unit: [parameter_name]}``.

    Examples
    --------
    >>> neclib.utils.quantity2builtin(
    ...     {"c": u.Quantity("299792458m/s")}, unit={"c": "km/s"}
    ... )
    {'c': 299792.458}
    >>> neclib.utils.quantity2builtin(
    ...     {"c": u.Quantity("299792458m/s")}, unit={"km/s": ["c"]}
    ... )
    {'c': 299792.458}

    Notes
    -----
    If there are parameters not in ``unit``, the values won't be converted.

    """
    _unit = {}
    for k, v in unit.items():
        if isinstance(v, (str, u.UnitBase)):
            _unit[k] = v
        else:
            _unit.update({name: k for name in v})

    def _convert(param, unit):
        return param if (unit is None) else param.to_value(unit)

    return {key: _convert(quantity[key], _unit.get(key, None)) for key in quantity}


def dAz2dx(
    az: Union[Number, u.Quantity], el: Union[Number, u.Quantity], unit: UnitType
) -> Number:
    az, el = list(
        map(lambda z: z.to_value(unit) if isinstance(z, u.Quantity) else z, [az, el])
    )
    rad = angle_conversion_factor(str(unit), "rad")  # type: ignore
    return az * math.cos(el * rad)  # type: ignore


def dx2dAz(
    x: Union[Number, u.Quantity], el: Union[Number, u.Quantity], unit: UnitType
) -> Number:
    x, el = list(
        map(lambda z: z.to_value(unit) if isinstance(z, u.Quantity) else z, [x, el])
    )
    rad = angle_conversion_factor(str(unit), "rad")  # type: ignore
    return x / math.cos(el * rad)  # type: ignore


class _GetQuantity:
    def __init__(self, default_unit=None):
        self._default_unit = default_unit
        self._nan = u.Quantity(float("nan"), self._default_unit)

    def __set_name__(self, owner, name):
        self._name = "_" + name

    def __set__(self, instance, value):
        if isinstance(value, str):
            value = u.Quantity(value)
        elif isinstance(value, (int, float)):
            value = u.Quantity(value, self._default_unit)
        setattr(instance, self._name, value)

    def __get__(self, instance, type=None):
        if instance is None:
            return self._nan
        return getattr(instance, self._name, self._nan)


@overload
def get_quantity(*, default_unit: Optional[UnitType]) -> _GetQuantity:
    ...


@overload
def get_quantity(
    *value: Union[str, Number, u.Quantity], unit: Optional[UnitType]
) -> u.Quantity:
    ...


def get_quantity(*value, unit=None, default_unit=None):
    """Convert values to quantity.

    Attention
    ---------
    This function has multiple signatures, one is simple converter from float values to
    Quantity, the other is attribute type validator. See examples.

    Parameters
    ----------
    value
        Values to convert.
    unit
        Unit to employ.
    default_unit
        Default unit to use.

    Examples
    --------
    To use as a type converter:
    >>> neclib.utils.get_quantity(1, 2, 3, unit="m")
    <Quantity [1., 2., 3.] m>
    >>> neclib.utils.get_quantity("1m", "2m", "3m")
    <Quantity [1., 2., 3.] m>

    To use as a type validator with default conversion:
    >>> class Test:
    ...     a = neclib.utils.get_quantity(default_unit="m")
    ...     b = neclib.utils.get_quantity(default_unit="m")
    >>> test = Test()
    >>> test.a = 1
    >>> test.b = "2m"
    >>> test.a
    <Quantity 1. m>
    >>> test.b
    <Quantity 2. m>

    """
    if not value:
        return _GetQuantity(default_unit)
    unit = None if unit is None else u.Unit(unit)

    def parser(v):
        if isinstance(v, u.Quantity):
            return v if unit is None else v.to(unit)
        return u.Quantity(v, unit)

    if len(value) == 1:
        return parser(*value)
    parsed = tuple(parser(v) for v in value)
    try:
        return u.Quantity(parsed)
    except (u.UnitConversionError, TypeError, ValueError):
        return parsed



import astropy.units as u
import ogameasure

from .thermometer_base import Thermometer


class Model218(Thermometer):

    Manufacturer = "LakeShore"
    Model = "Model218"

    Identifier = "host"

    def __init__(self) -> None:
        self.thermometer = ogameasure.Lakeshore.model218_usb(self.Config.usb_port)

    def get_temp(self) -> u.Quantity:
        data = self.thermometer.kelvin_reading_query()
        return data * u.K

    def finalize(self) -> None:
        self.thermometer.com.close()



"""Informative exception, regarding the NECST system status."""

__all__ = ["NECSTAuthorityError", "NECSTConfigurationError", "NECSTTimeoutError"]


class NECSTAuthorityError(Exception):
    """Error related to controlling priority.

    Examples
    --------
    >>> raise neclib.NECSTAuthorityError("reason or advise to fix this error")

    """

    pass


class NECSTConfigurationError(Exception):
    """Error related to parameter configuration."""

    pass


ConfigurationError = NECSTConfigurationError
"""Alias for :class:`NECSTConfigurationError`, for backward compatibility."""


class NECSTTimeoutError(Exception):
    """Error related to communication timeout."""

    pass



from abc import abstractmethod

import astropy.units as u

from ..device_base import DeviceBase


class Attenuator(DeviceBase):
    @abstractmethod
    def set_level(self, level_dB: int, ch: int):
        ...

    @abstractmethod
    def get_level(self, ch: int) -> u.Quantity:
        ...



import time
from typing import Optional, Tuple

import numpy as np


class Resize:
    def __init__(self, keep_duration_sec: float = 1.0) -> None:
        self.data_list = []
        self.keep_duration = keep_duration_sec

    def push(self, data: list, timestamp: Optional[float] = None) -> None:
        timestamp = time.time() if timestamp is None else timestamp
        self.data_list.append((data, timestamp))
        self._discard_outdated_data()

    def _discard_outdated_data(self) -> None:
        if len(self.data_list) > 1:
            while self.data_list[0][1] < time.time() - self.keep_duration:
                self.data_list.pop(0)

    def get(self, range: Tuple[int, int], n_samples: Optional[int] = None) -> list:
        """Cut spectral data with arbitrary range.

        Parameter
        ---------
        range
            The first and last index of spectral data you want (last index won't be
            included).
        n_samples
            The number of (coarse) samples you want.

        Example
        -------
        >>> spectral_data = neclib.data.Resize()
        >>> spectral_data.push(data)
        Repeat several times ...
        >>> spectral_data.get([8192, 16384], 100)

        """
        self._discard_outdated_data()
        spec_array = np.array([data[0] for data in self.data_list])

        cut_spec_array = spec_array[:, slice(*range)].mean(axis=0)
        # Validity of taking mean at this point (before interpolation) isn't checked,
        # just for simple implementation using `np.interp`.

        if n_samples is not None:
            target_idx = np.linspace(0, len(cut_spec_array), n_samples)
            original_idx = np.linspace(0, len(cut_spec_array), len(cut_spec_array))
            cut_spec_array = np.interp(target_idx, original_idx, cut_spec_array)
        return cut_spec_array.tolist()



from ...utils import busy
from .bias_setter_base import BiasSetter


class CPZ340816(BiasSetter):

    Manufacturer = "Interface"
    Model = "CPZ340816"

    Identifier = "rsw_id"

    def __init__(self):
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.param_buff = {
            1: 0.0,
            2: 0.0,
            3: 0.0,
            4: 0.0,
            5: 0.0,
            6: 0.0,
            7: 0.0,
            8: 0.0,
            9: 0.0,
            10: 0.0,
            11: 0.0,
            12: 0.0,
            13: 0.0,
            14: 0.0,
            15: 0.0,
            16: 0.0,
        }  # All in [V]
        self.da = pyinterface.open(3408, self.rsw_id)

    def set_voltage(self, voltage_mV: float, ch: int) -> None:
        if ch not in self.param_buff.keys():
            raise ValueError("Invaild channel {ch}")
        else:
            self.param_buff[ch] = voltage_mV / 3

    def output_voltage(self) -> None:
        with busy(self, "busy"):
            for i in range(0, 16):
                ch = int(list(self.param_buff.keys())[i])
                voltage = list(self.param_buff.values())[i]

                self.da.output_da(ch, voltage)

                continue

    def finalize(self) -> None:
        self.da.finalize()
        self.param_buff = {
            1: 0.0,
            2: 0.0,
            3: 0.0,
            4: 0.0,
            5: 0.0,
            6: 0.0,
            7: 0.0,
            8: 0.0,
            9: 0.0,
            10: 0.0,
            11: 0.0,
            12: 0.0,
            13: 0.0,
            14: 0.0,
            15: 0.0,
            16: 0.0,
        }



from abc import abstractmethod

import astropy.units as u

from ..device_base import DeviceBase


class WeatherStation(DeviceBase):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def get_temp(self) -> u.Quantity:
        ...

    @abstractmethod
    def get_humid(self) -> float:
        ...

    @abstractmethod
    def get_press(self) -> u.Quantity:
        ...

    @abstractmethod
    def finalize(self) -> None:
        ...



from abc import abstractmethod
import astropy.units as u

from ..device_base import DeviceBase


class SignalGenerator(DeviceBase):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def set_freq(self, freq_GHz) -> None:
        ...

    @abstractmethod
    def set_power(self, power_dBm) -> None:
        ...

    @abstractmethod
    def get_freq(self) -> u.Quantity:
        ...

    @abstractmethod
    def get_power(self) -> u.Quantity:
        ...

    @abstractmethod
    def start_output(self) -> None:
        ...

    @abstractmethod
    def stop_output(self) -> None:
        ...

    @abstractmethod
    def get_output_status(self) -> bool:
        ...

    @abstractmethod
    def finalize(self) -> None:
        ...



import astropy.units as u
import ogameasure

from ... import get_logger, utils
from .power_meter_base import PowerMeter


class MA24126A(PowerMeter):

    Model = "MA24126A"
    Manufacturer = "Anritsu"

    Identifier = "port"

    def __init__(self):
        self.logger = get_logger(self.__class__.__name__)
        self.io = ogameasure.Anritsu.ma24126a(self.Config.port)
        self.io.start()

    def get_power(self) -> u.Quantity:
        with utils.busy(self, "busy"):
            ret = self.io.power()
            power = float(ret.decode().split("\n")[0])
            return power * u.mW

    def zero_set(self) -> None:
        with utils.busy(self, "busy"):
            self.logger.info("##### usb power meter is doing zero setting now ####")
            self.io.zero_set()
            self.logger.info("##### usb power meter finished zero setting  ####")

    def finalize(self):
        self.io.close()



__all__ = ["CoordCalculator"]

import os
import time
from typing import List, Optional, Tuple, TypeVar, Union

import astropy.constants as const
import astropy.units as u
import numpy as np
from astropy.coordinates import (
    AltAz,
    BaseCoordinateFrame,
    EarthLocation,
    SkyCoord,
    get_body,
)
from astropy.time import Time

from .. import config, get_logger, utils
from ..parameters.pointing_error import PointingError
from ..typing import Number, UnitType
from .frame import parse_frame

T = TypeVar("T", Number, u.Quantity)


class CoordCalculator:
    """Calculate horizontal coordinate.

    Parameters
    ----------
    location
        Location of observatory.
    pointing_param_path
        Path to pointing parameter file.
    pressure
        Atmospheric pressure at the observatory.
    temperature
        Temperature at the observatory.
    relative_humidity
        Relative humidity at the observatory.
    obswl
        Observing wavelength.
    obsfreq
        Observing frequency of EM-wave, to compute diffraction correction.

    Attributes
    ----------
    location: EarthLocation
        Location of observatory.
    pressure: Quantity
        Atmospheric pressure, to compute diffraction correction. If dimensionless value
        is given, it is assumed to be in ``hPa``.
    temperature: Quantity
        Temperature, to compute diffraction correction. If dimensionless value is given,
        it is assumed to be in ``K``.
    relative_humidity: Quantity or float
        Relative humidity, to compute diffraction correction.
    obswl: Quantity
        Observing wavelength, to compute diffraction correction. If dimensionless value
        is given, it is assumed to be in ``m``.

    Examples
    --------
    >>> location = EarthLocation("138.472153deg", "35.940874deg", "1386m")
    >>> path = "path/to/pointing_param.toml"
    >>> pressure = 850 << u.hPa
    >>> temperature = 300 << u.K
    >>> humid = 0.30
    >>> obswl = 230.5 << u.GHz
    >>> calculator = neclib.coordinates.CoordCalculator(
    ...     location, path, pressure=pressure, temperature=temperature,
    ...     relative_humidity=humid, obswl=obswl)

    """

    pressure = utils.get_quantity(default_unit="hPa")
    temperature = utils.get_quantity(default_unit="K")
    relative_humidity = utils.get_quantity(default_unit="")
    obswl = utils.get_quantity(default_unit="m")

    def __init__(
        self,
        location: EarthLocation,
        pointing_param_path: Optional[os.PathLike] = None,
        *,
        pressure: Optional[u.Quantity] = None,
        temperature: Optional[u.Quantity] = None,
        relative_humidity: Optional[Union[Number, u.Quantity]] = None,
        obswl: Optional[u.Quantity] = None,
        obsfreq: Optional[u.Quantity] = None,
    ) -> None:
        self.logger = get_logger(self.__class__.__name__)

        if (obswl is not None) and (obsfreq is not None):
            if obswl != const.c / obsfreq:  # type: ignore
                raise ValueError("Specify ``obswl`` or ``obs_freq``, not both.")

        self.location = location
        self.pointing_param_path = pointing_param_path
        self.pressure = pressure
        self.temperature = temperature
        self.relative_humidity = relative_humidity

        if obsfreq is not None:
            self.obswl = const.c / obsfreq  # type: ignore

        if pointing_param_path is not None:
            self.pointing_error_corrector = PointingError.from_file(pointing_param_path)
        else:
            self.logger.warning("Pointing error correction is disabled.")
            dummy = PointingError()
            dummy.refracted2apparent = lambda az, el: (az, el)  # type: ignore
            self.pointing_error_corrector = dummy

        diffraction_params = ["pressure", "temperature", "relative_humidity", "obswl"]
        not_set = list(filter(lambda x: getattr(self, x) is None, diffraction_params))
        if len(not_set) > 0:
            self.logger.warning(
                f"{not_set} are not given. Diffraction correction is disabled."
            )

    def _get_altaz_frame(self, obstime: Union[Number, Time]) -> AltAz:
        obstime = self._convert_obstime(obstime)
        if self.temperature is not None:
            self.temperature = self.temperature.to(
                "deg_C", equivalencies=u.temperature()
            )
        return AltAz(
            obstime=obstime,
            location=self.location,
            pressure=self.pressure,
            temperature=self.temperature,
            relative_humidity=self.relative_humidity,
            obswl=self.obswl,
        )

    def _convert_obstime(self, obstime: Union[Number, Time, None]) -> Time:
        if obstime is None:
            obstime = self._auto_schedule_obstime()
        return obstime if isinstance(obstime, Time) else Time(obstime, format="unix")

    def _auto_schedule_obstime(self, duration=1):
        """Automatically generate sequence of time."""
        now = time.time()
        frequency = config.antenna_command_frequency
        offset = config.antenna_command_offset_sec
        return Time(
            [now + offset + i / frequency for i in range(int(frequency * duration))],
            format="unix",
        )

    def get_altaz_by_name(
        self,
        name: str,
        obstime: Optional[Union[Number, Time]] = None,
    ) -> Tuple[u.Quantity, u.Quantity, List[float]]:
        """天体名から地平座標 az, el(alt) を取得する

        Parameters
        ----------
        name
            Name of celestial object.
        obstime
            Time the observation is done.

        Examples
        --------
        >>> calculator.get_altaz_by_name("M42", time.time())
        <SkyCoord (az, alt) in deg (274.55435678, -15.3762009)>

        """
        obstime = self._convert_obstime(obstime)
        try:
            coord = get_body(name, obstime)
        except KeyError:
            coord = SkyCoord.from_name(name)
        altaz = coord.transform_to(self._get_altaz_frame(obstime))
        return (
            *self.pointing_error_corrector.refracted2apparent(
                altaz.az, altaz.alt  # type: ignore
            ),
            obstime.unix,
        )

    def get_altaz(
        self,
        lon: T,  # 変換前の経度
        lat: T,  # 変換前の緯度
        frame: Union[str, BaseCoordinateFrame],  # 変換前の座標系
        *,
        unit: Optional[UnitType] = None,
        obstime: Optional[Union[Number, Time]] = None,
    ) -> Tuple[u.Quantity, u.Quantity, List[float]]:
        """Get horizontal coordinate from longitude and latitude in arbitrary frame.

        Parameters
        ----------
        lon
            Longitude of target.
        lat
            Latitude of target.
        frame
            Coordinate frame, in which ``lon`` and ``lat`` are given.
        unit
            Angular unit in which ``lon`` and ``lat`` are given. If they are given as
            ``Quantity``, this parameter will be ignored.
        obstime
            Time the observation is done.

        Examples
        --------
        >>> calculator.get_altaz(30 << u.deg, 45 << u.deg, "fk5", obstime=time.time())
        <SkyCoord (az, alt) in deg (344.21675916, -6.43235393)>

        """
        obstime = self._convert_obstime(obstime)
        lon, lat = utils.get_quantity(lon, lat, unit=unit)  # type: ignore
        if getattr(frame, "name", frame) == "altaz":
            frame = self._get_altaz_frame(time.time())
            (
                apparent_az,
                apparent_alt,
            ) = self.pointing_error_corrector.refracted2apparent(lon, lat)
            return (
                np.broadcast_to(apparent_az, obstime.shape) << apparent_az.unit,
                np.broadcast_to(apparent_alt, obstime.shape) << apparent_alt.unit,
                obstime.unix,
            )
        elif isinstance(frame, str):
            frame = parse_frame(frame)

        altaz = SkyCoord(lon, lat, frame=frame).transform_to(
            self._get_altaz_frame(obstime)
        )
        return (
            *self.pointing_error_corrector.refracted2apparent(
                altaz.az, altaz.alt  # type: ignore
            ),
            obstime.unix,
        )



"""Utility functions for data structure handling."""

__all__ = [
    "ParameterList",
    "AzElData",
    "ParameterMapping",
    "ValueRange",
    "toCamelCase",
    "to_snake_case",
    "AliasedDict",
]

import re
from collections import UserDict
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Generic,
    Hashable,
    Iterable,
    Iterator,
    Literal,
    Optional,
    TypeVar,
)

import numpy as np

from ..typing import SupportsComparison


class ParameterList(list):
    """List, specialized in parameter storing.

    Parameters
    ----------
    value
        Iterable object to be converted to ParameterList.

    Examples
    --------
    >>> neclib.utils.ParameterList([0, 0])
    ParameterList([0, 0])
    >>> neclib.utils.ParameterList(range(5))
    ParameterList([0, 1, 2, 3, 4])

    """

    def __init__(self, value: Iterable = []):
        super().__init__(value)

    @classmethod
    def new(cls, length: int, initvalue: Any = float("nan")) -> "ParameterList":
        """Create new ParameterList instance filled with initial value.

        Parameters
        ----------
        length
            Length of ``ParameterList`` to be created.
        initvalue
            Initial value to fill the ``ParameterList``.

        Examples
        --------
        >>> neclib.utils.ParameterList.new(3, 100)
        ParameterList([100, 100, 100])

        """
        return cls([initvalue for _ in range(length)])

    def push(self, value: Any) -> None:
        """Append new value to ParameterList, preserving its length.

        Parameters
        ----------
        value
            Value to be appended.

        Examples
        --------
        >>> param = neclib.utils.ParameterList([1, 2])
        >>> param.push(5)
        >>> param
        ParameterList([2, 5])

        """
        self.append(value)
        self.pop(0)

    def copy(self):
        """Return copied ParameterList.

        Examples
        --------
        >>> param = neclib.utils.ParameterList([1, 2])
        >>> param.copy()
        ParameterList([1, 2])

        """
        return self.__class__(super().copy())

    def map(self, func: Callable) -> "ParameterList":
        """Map a function to every element in the ParameterList.

        Parameters
        ----------
        func
            Function to apply.

        Examples
        --------
        >>> param = neclib.utils.ParameterList([1, 2])
        >>> param.map(lambda x: 10 * x)
        ParameterList([10, 20])

        """
        return self.__class__(list(map(func, self)))

    def __repr__(self) -> str:
        return self.__class__.__name__ + "(" + super().__repr__() + ")"


@dataclass
class AzElData:
    az: Any = None
    """Parameter related to azimuth axis."""
    el: Any = None
    """Parameter related to elevation axis."""


class ParameterMapping(dict):
    """Dict, with attribute access to parameter supported.

    Parameter mapping, supports both dict-key syntax and attribute access syntax. Dict
    methods are fully supported.

    Examples
    --------
    >>> param = neclib.utils.ParameterMapping(a=1, b=2)
    >>> param["a"]
    1
    >>> param.a
    1
    >>> neclib.utils.ParameterMapping({"a": 1, "b": 2}) == param
    True

    """

    def __repr__(self) -> str:
        return self.__class__.__name__ + "(" + super().__repr__() + ")"

    def __getattr__(self, name: Hashable) -> Any:
        try:
            return self[name]
        except KeyError as e:  # Raise AttributeError instead of KeyError.
            raise AttributeError(f"No attribute '{name}'") from e

    def copy(self):
        """Return copied ParameterMapping.

        Examples
        --------
        >>> param = neclib.utils.ParameterMapping(a=1, b=2)
        >>> param.copy()
        ParameterMapping({'a': 1, 'b': 2})

        """
        return self.__class__(super().copy())


T = TypeVar("T", bound=SupportsComparison)


class ValueRange(Generic[T]):
    """Utility type for value range checking.

    Parameters
    ----------
    lower
        Lower bound of the range. Any type with comparison support is allowed.
    upper
        Upper bound of the range. Any type with comparison support is allowed.
    strict
        If ``True``, the value exactly equals to the bound will judged to be
        not in the range.

    Examples
    --------
    >>> valid_value = neclib.utils.ValueRange(0, 1)
    >>> 0.5 in valid_value
    True
    >>> -1 in valid_value
    False

    >>> valid_str = neclib.utils.ValueRange("aaa", "bbb")
    >>> "abc" in valid_str
    True

    >>> _ = [print(limits) for limits in valid_value]
    aaa
    bbb

    """

    def __init__(self, lower: T, upper: T, strict: bool = False) -> None:
        try:
            if lower > upper:
                raise ValueError("Lower bound must be smaller than upper bound.")
        except TypeError:
            raise TypeError("Bounds must support comparison.")

        self.lower, self.upper = lower, upper
        self.strict = strict

    def __contains__(self, value: Any) -> bool:
        if self.strict:
            return self.lower < value < self.upper
        return self.lower <= value <= self.upper

    def __iter__(self) -> Iterator[T]:
        return iter((self.lower, self.upper))

    def __repr__(self) -> str:
        class_name = self.__class__.__name__
        return f"{class_name}({self.lower}, {self.upper}, strict={self.strict})"

    def __eq__(self, other) -> bool:
        if not isinstance(other, self.__class__):
            return False
        eq_limits = (self.lower == other.lower) and (self.upper == other.upper)
        return eq_limits and self.strict is other.strict

    @property
    def width(self) -> Optional[T]:
        """Width of the range.

        Examples
        --------
        >>> valid_value = neclib.utils.ValueRange(0, 1)
        >>> valid_value.width
        1

        """
        try:
            return self.upper - self.lower  # type: ignore
        except TypeError:
            return None

    def contain_all(self, values: Iterable[Any]) -> bool:
        """Check if all values are in the range.

        Parameters
        ----------
        values
            Iterable object to be checked.

        Examples
        --------
        >>> valid_value = neclib.utils.ValueRange(0, 1)
        >>> valid_value.contain_all([0.5, 1.6])
        False

        """
        values = np.asanyarray(values)
        if self.strict:
            ret = ((self.lower < values) & (values < self.upper)).all()
        else:
            ret = ((self.lower <= values) & (values <= self.upper)).all()
        return bool(ret)

    def contain_any(self, values: Iterable[Any]) -> bool:
        """Check if any value is in the range.

        Parameters
        ----------
        values
            Iterable object to be checked.

        Examples
        --------
        >>> valid_value = neclib.utils.ValueRange(0, 1)
        >>> valid_value.contain_any([0.5, 1.6])
        True

        """
        values = np.asanyarray(values)
        if self.strict:
            ret = ((self.lower < values) & (values < self.upper)).any()
        else:
            ret = ((self.lower <= values) & (values <= self.upper)).any()
        return bool(ret)

    def map(self, func: Callable[[T], Any]) -> "ValueRange":
        """Map a function to upper and lower bounds.

        Parameters
        ----------
        func
            Function to apply.

        Examples
        --------
        >>> valid_value = neclib.utils.ValueRange(0, 1)
        >>> valid_value.map(lambda x: 10 * x)
        ValueRange(0, 10, strict=False)

        """
        return self.__class__(func(self.lower), func(self.upper), self.strict)


def toCamelCase(
    data: str,
    kind: Literal["upper", "pascal", "bumpy", "python", "lower", ""] = "python",
) -> str:
    PascalCase = re.sub(
        r"([A-Za-z])([a-z0-9]*?)(_|$|\s)",
        lambda mo: mo.group(1).upper() + mo.group(2),
        data,
    )
    if kind.lower() in ["upper", "pascal", "bumpy", "python"]:
        return PascalCase
    elif kind.lower() in ["lower", ""]:
        return re.sub(
            r"^([A-Z])([a-z0-9]+?)",
            lambda mo: mo.group(1).lower() + mo.group(2),
            PascalCase,
        )
    raise ValueError(f"Unknown kind: {kind}")


def to_snake_case(data: str) -> str:
    ret = re.sub(r"(?!^)([A-Z])([a-z]+)", r"_\1\2", data).lower()
    return re.sub(r"\s", "_", ret)


class AliasedDict(UserDict):
    def __init__(self, *args, **kwargs):
        self.__aliases = {}
        super().__init__(*args, **kwargs)

    def alias(self, **kwargs) -> None:
        for newkey, existingkey in kwargs.items():
            if newkey in self.keys():
                raise KeyError(
                    f"Key {newkey!r} already exists, cannot create alias with the name"
                )
            self.__aliases[newkey] = existingkey

    def __setitem__(self, key, value):
        if key in self.__aliases.values():
            alias_key = (k for k, v in self.__aliases.items() if v == key)
            del self.__aliases[next(alias_key)]
        super().__setitem__(key, value)

    def __missing__(self, key):
        try:
            return self[self.__aliases[key]]
        except KeyError:
            raise KeyError(key) from None



from types import ModuleType
from typing import Dict, List, Type

from .. import config, get_logger, utils
from ..exceptions import NECSTConfigurationError
from .device_base import DeviceBase

logger = get_logger(__name__)


def parse_device_configuration(
    modules: List[ModuleType],
) -> Dict[str, Type[DeviceBase]]:
    implementations = list_implementations(modules)

    devices = config.dev
    if devices is None:
        return {}

    parsed: Dict[str, Type[DeviceBase]] = {}
    for k, v in devices.items():
        if v.lower() in implementations.keys():
            impl = implementations[v.lower()]
            new: Type[DeviceBase] = type(
                utils.toCamelCase(k), (impl,), {}  # type: ignore
            )

            parsed[k] = parsed[utils.toCamelCase(k)] = new
            try:  # Run __new__ once to ensure all configurations set.
                new.__new__(new)
            except Exception as e:
                logger.warning(f"Failed to initialize device {k}: {e}")
        else:
            raise NECSTConfigurationError(
                f"Driver implementation for device '{v}' ({k}) not found."
            )
    return parsed


def list_implementations(modules: List[ModuleType]) -> Dict[str, Type[DeviceBase]]:
    def list_implementations_single_module(
        module: ModuleType,
    ) -> Dict[str, Type[DeviceBase]]:
        impl: Dict[str, Type[DeviceBase]] = {}
        for attrname in dir(module):
            attr = getattr(module, attrname)
            try:
                if issubclass(attr, DeviceBase):
                    impl[attrname.lower()] = attr
            except TypeError:
                del attr

        return impl

    implementations: Dict[str, Type[DeviceBase]] = {}
    for module in modules:
        impl = list_implementations_single_module(module)
        implementations.update(impl)
    if len(implementations) != len(set(implementations.keys())):
        raise ValueError("Implemented device model isn't unique.")
    return implementations



from abc import ABC, abstractmethod
from typing import Any, Dict, Final, Optional, Type, final

from .. import NECSTConfigurationError, utils
from ..configuration import Configuration, _Cfg, config


class _CfgManager:
    def __set_name__(self, owner: Type[object], name: str) -> None:
        self.name = "_" + name

    def __get__(self, instance: object, type: Optional[Type[object]] = None):
        if instance is None:
            return None
        return getattr(instance, self.name, None)

    def __set__(self, instance: object, value: Any) -> None:
        if not isinstance(value, _Cfg):
            raise TypeError("Configuration object expected.")

        current_config = getattr(instance, self.name, None)

        if current_config < value:
            # If new configuration is the superset of currently assigned one, update it.
            setattr(instance, self.name, value)
        elif (current_config is None) and (value is None):
            pass
        elif current_config >= value:
            pass
        else:
            raise NECSTConfigurationError(
                "Currently this software doesn't support merging multiple config "
                "objects. Please include all configurations for single device in single"
                " config object (others should contain subset of that; notably the "
                "identifier) or make identical config objects for all aliases."
            )


class DeviceBase(ABC):
    """Base class for all devices.

    Parameters
    ----------
    config
        Device configuration, as ``Configuration`` group object.
    key
        Identifier to uniquely specify single device.

    """

    _instances: Final[Dict[str, Any]] = {}
    conf: Configuration

    Manufacturer: str = ""
    Model: str

    Identifier: Optional[str] = None
    Config = _CfgManager()

    @final
    def __new__(cls):
        if cls.Identifier is None:
            return super().__new__(cls)

        _cfg = getattr(config, utils.to_snake_case(cls.__name__))

        identifier = (
            "" if cls.Identifier is None else getattr(_cfg, cls.Identifier, None)
        )
        key = f"{cls.Model}_{cls.Manufacturer}_{identifier}"
        if key not in cls._instances:
            cls._instances[key] = super().__new__(cls)

        cls._instances[key].Config = _cfg
        if cls._instances[key].__class__ is not cls:
            cls._instances[key].__init__()
        return cls._instances[key]

    @final
    def __repr__(self) -> str:
        model = f"model={self.Model!r}"
        manufacturer = f"manufacturer={self.Manufacturer!r}"
        ident = ""
        if self.Identifier is not None:
            ident = f"{self.Identifier}={getattr(self.Config, self.Identifier, None)}"
        return f"{self.__class__.__name__}({model}, {manufacturer}, {ident})"

    @final
    def __str__(self) -> str:
        return self.__repr__()

    @abstractmethod
    def finalize(self) -> None:
        """Safely shut down the device.

        This method should always be called on quitting the program.

        """
        ...



__all__ = ["CPZ7415V"]

import time
from typing import Dict, Literal

import astropy.units as u

from ... import get_logger, utils
from .motor_base import Motor


class CPZ7415V(Motor):
    """Pulse controller, which can handle up to 4 motors (axes).

    Notes
    -----
    Configuration items for this device:

    rsw_id : {0, 1, ..., 16} or {"0", "1", ..., "9", "A", ..., "F"}
        Board identifier. This should be set to the same value as the rotary switch
        "RSW1" mounted on the side of the board. The board is shipped with default RSW1
        setting of 0. This ID would be non-zero, when multiple PCI board of same model
        are mounted on a single FA (Factory Automation) controller.
    useaxes : {"x", "y", "z", "u"} or concatenation of them
        Axes to control.
    {axis}_mode : {"jog", "org", "ptp", "timer", ...}
        Motion mode for the axis. ``"jog"`` for speed control, ``"ptp"`` for absolute
        position control.
    {axis}_pulse_conf : Dict[str, int]
        Keys should be ["PULSE", "OUT", "DIR", "WAIT", "DUTY"], and values should be 1
        or 0. If "PULSE" is 1, 2-pulse mode is used. If "OUT" is 1, output pulse is
        active-high. If "DIR" is 1, direction output is high when CCW motion (1-pulse
        mode), direction output is always low (2-pulse mode). If "WAIT" is 1, delay of
        200us is inserted on the direction change. If "DUTY" is 1, duty cycle is
        adaptively changed according to the speed, otherwise always 50%.
    {axis}_motion_clock : int
        Pulse frequency scale factor, valid values are 2 through 4095. The frequency is
        calculated by ``300 / (clock + 1)``.
    {axis}_motion_acc_mode : {"acc_normal", "acc_sin"}
        Acceleration mode. ``"acc_normal"`` for linear acceleration, ``"acc_sin"``
        for S-curve acceleration.
    {axis}_motion_low_speed : int
        Startup speed, valid values are 1 through 65535.
    {axis}_motion_speed : int
        Steady state speed, valid values are 1 through 65535. This will be used in
        "ptp" motion.
    {axis}_motion_acc : int
        Acceleration time, valid values are 1 through 65535.
    {axis}_motion_dec : int
        Deceleration time, valid values are 1 through 65535.
    {axis}_motion_step : int
        Number of pulses, used in "ptp" (valid values are -134217728 through 134217727),
        "org_search", "timer" (valid values are 1 through 134217727) motion. This
        setting is ignored in other modes.
    axis_{alias} : {"x", "y", "z", "u"}
        Mapping from the controller axes to telescope control axes. The ``{alias}``
        should be ["az", "el"].
    speed_to_pulse_factor_{axis} : float
        Conversion factor from speed to pulse frequency. This includes the gear ratio.
        For axes which won't be controlled in speed, there's no need to define this
        parameter for them.

    All parameters prefixed with ``{axis}_`` need to be defined for each axis in
    ``useaxes``.
    See http://www.interface.co.jp/download/tutorial/tut0053_14.pdf for details.

    """

    Manufacturer = "Interface"
    Model = "CPZ7415V"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)

        self.rsw_id = self.Config.rsw_id
        self.use_axes = self.Config.useaxes.lower()
        self.axis_mapping = dict(self.Config.axis.items())
        self.speed_to_pulse_factor = utils.AliasedDict(
            self.Config.speed_to_pulse_factor.items()
        )
        _config = {ax: getattr(self.Config, ax) for ax in self.use_axes}

        self.speed_to_pulse_factor.alias(**{v: k for k, v in self.axis_mapping.items()})

        self.motion = {
            ax: dict(getattr(self.Config, f"{ax}_motion").items())
            for ax in self.use_axes
        }
        self.motion_mode = {ax: conf.mode for ax, conf in _config.items()}
        self.pulse_conf = {ax: conf.pulse_conf for ax, conf in _config.items()}
        self.default_speed = {ax: conf.motion_speed for ax, conf in _config.items()}
        self.low_speed = {ax: conf.motion_low_speed for ax, conf in _config.items()}

        self.last_direction = {ax: 0 for ax in self.use_axes}

        self.io = self._initialize_io()

    @utils.skip_on_simulator
    def _initialize_io(self):
        import pyinterface

        io = pyinterface.open(7415, self.rsw_id)
        if io is None:
            raise RuntimeError("Cannot communicate with the PCI board.")

        count = [c if c != 0 else 1 for c in io.read_counter(self.use_axes, "counter")]
        io.initialize()

        # HACK: Escape from origin
        io.write_counter(self.use_axes, "counter", [1] * len(self.use_axes))
        io.write_counter(self.use_axes, "counter", count)

        for ax in self.use_axes:
            io.set_pulse_out(ax, "method", [self.pulse_conf[ax]])
        io.set_motion(self.use_axes, list(self.motion_mode.values()), self.motion)

        do = [int(self.motion_mode.get(ax, "") == "jog") for ax in "xyzu"]
        io.output_do(do)
        return io

    @property
    def current_motion(self) -> Dict[str, int]:
        status = map(int, self.io.driver.check_move_onoff(self.use_axes))
        return {ax: st for ax, st in zip(self.use_axes, status)}

    def _parse_ax(self, axis: str) -> Literal["x", "y", "z", "u"]:
        if axis in self.use_axes:
            return axis
        return self.axis_mapping[axis.lower()]

    def get_speed(self, axis: str) -> u.Quantity:
        ax = self._parse_ax(axis)
        with utils.busy(self, "_busy"):
            speed = self.io.read_speed(ax)[0]
            return speed / self.speed_to_pulse_factor[ax] * u.Unit("deg/s")

    def get_step(self, axis: str) -> int:
        ax = self._parse_ax(axis)
        with utils.busy(self, "_busy"):
            return self.io.read_counter(ax, cnt_mode="counter")[0]

    def set_speed(self, speed: float, axis: str) -> None:
        ax = self._parse_ax(axis)
        speed *= self.speed_to_pulse_factor[ax]

        if abs(speed) < self.low_speed[ax]:
            self._stop(ax)
            self.last_direction[ax] = 0
            return
        if self.motion_mode[ax] == "jog":
            if (self.last_direction[ax] * speed > 0) and (self.current_motion[ax] != 0):
                self._change_speed(abs(speed), ax)
            else:
                direction = 1 if speed > 0 else -1
                step = direction  # Not an absolute position.
                self.last_direction[ax] = direction
                self._start(abs(speed), step, ax)
                time.sleep(1e-2)

    def set_step(self, step: int, axis: str) -> None:
        ax = self._parse_ax(axis)
        if self.motion_mode[ax] != "ptp":
            raise RuntimeError(
                "Position setting is only supported in point-to-point (ptp) mode, "
                f"but {axis=!r} is controlled in {self.motion_mode[ax]!r} mode."
            )
        if self.current_motion[ax] != 0:
            self._change_step(step, ax)
        else:
            speed = self.motion[ax]["speed"]
            self._start(speed, step, ax)

    def _change_step(self, step: int, axis: Literal["x", "y", "z", "u"]) -> None:
        with utils.busy(self, "_busy"):
            self.io.change_step(axis=axis, step=[step])

    def _change_speed(self, speed: float, axis: Literal["x", "y", "z", "u"]) -> None:
        with utils.busy(self, "_busy"):
            self.io.change_speed(axis=axis, mode="accdec_change", speed=[speed])

    def _start(
        self, speed: float, step: int, axis: Literal["x", "y", "z", "u"]
    ) -> None:
        self._stop(axis)
        with utils.busy(self, "_busy"):
            self.motion[axis]["speed"] = speed
            self.motion[axis]["step"] = int(step)

            axis_mode = [self.motion_mode[axis]]
            self.io.set_motion(axis=axis, mode=axis_mode, motion=self.motion)
            self.io.start_motion(
                axis=axis, start_mode="const", move_mode=self.motion_mode[axis]
            )

    def _stop(self, axis: Literal["x", "y", "z", "u"]) -> None:
        with utils.busy(self, "_busy"):
            self.io.stop_motion(axis=axis, stop_mode="immediate_stop")
            while int(self.io.driver.get_main_status(axis)[0][0]) != 0:
                time.sleep(1e-4)

    def finalize(self) -> None:
        self.io.output_do([0, 0, 0, 0])
        for ax in self.use_axes:
            self._stop(ax)



from abc import abstractmethod

from ..device_base import DeviceBase


class BiasSetter(DeviceBase):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def set_voltage(self, voltage_mV: float, ch: int) -> None:
        ...

    @abstractmethod
    def output_voltage(self) -> None:
        ...

    @abstractmethod
    def finalize(self) -> None:
        ...



__all__ = ["DriveLimitChecker"]

from typing import Optional

import astropy.units as u
import numpy as np

from .. import get_logger, utils
from ..typing import QuantityValue, UnitType


class DriveLimitChecker:
    """Find optimum and safe angle to drive to.

    The optimization includes:

    1. Apply drive range constraint
    2. Avoid direction unwrapping (360deg drive) during observation
    3. Avoid over-180deg drive

    Parameters
    ----------
    limit
        Drive range limit, the optimization result never go out of this range.
    preferred_limit
        Drive range limit, which can be violated to continue observation or no other
        choice found.
    unit
        Angular unit in which the parameters are given.
    max_observation_size
        If separation between current and target coordinates is smaller than this value,
        direction unwrapping won't occur, possibly violating ``preferred_limit``, to
        not interrupt observation.

    Examples
    --------
    >>> checker = neclib.coordinates.DriveLimitChecker(
    ...     [-260 << u.deg, 260 << u.deg], [-250 << u.deg, 250 << u.deg]
    ... )
    >>> checker.optimize(current=-200 << u.deg, target=170 << u.deg)
    <Quantity -190. deg>

    """

    def __init__(
        self,
        limit: utils.ValueRange[QuantityValue],  # type: ignore
        preferred_limit: Optional[utils.ValueRange[QuantityValue]] = None,  # type: ignore  # noqa: E501
        *,
        unit: Optional[UnitType] = None,
        max_observation_size: QuantityValue = 5 << u.deg,  # type: ignore
    ) -> None:
        self.logger = get_logger(self.__class__.__name__)

        limit = utils.get_quantity(*limit, unit=unit)  # type: ignore
        if preferred_limit is None:
            preferred_limit = limit
        preferred_limit = utils.get_quantity(
            *preferred_limit, unit=unit  # type: ignore
        )
        max_observation_size = utils.get_quantity(max_observation_size, unit=unit)

        self.limit = utils.ValueRange(*limit)  # type: ignore
        self.preferred_limit = utils.ValueRange(*preferred_limit)  # type: ignore
        self.max_observation_size = max_observation_size

        # This script doesn't silently swap the lower and upper bounds for `limit` and
        # `preferred_limit`, since the ambiguity can cause destructive result, i.e.
        # [270, 90]deg can mean [-90, 90]deg, which is opposite to [90, 270]deg.

    def optimize(
        self,
        current: QuantityValue,
        target: QuantityValue,
        unit: Optional[UnitType] = None,
    ) -> Optional[u.Quantity]:
        """Optimize the coordinate to command.

        Parameters
        ----------
        current
            Current coordinate, scalar or n-d array of shape [..., N].
        target
            Target coordinate to be optimized, scalar or n-d array of shape [..., N] or
            [..., N, T], where T is time series length.
        unit
            Angular unit in which the ``current`` and ``target`` are given.

        """
        _current, _target = utils.get_quantity(current, target, unit=unit)
        target_shape = _target.shape

        if _current.shape == target_shape:
            _target = _target[..., None]  # Add time series dimension
        elif _current.shape == target_shape[:-1]:
            pass
        else:
            shape = _current.shape
            raise ValueError(
                f"Target shape should be ({shape}) or ({shape}, T), got {target_shape}"
            )

        _360deg = 360 << u.deg  # type: ignore
        min_candidate = _target - _360deg * ((_target - self.limit.lower) // _360deg)

        optimum = []
        for i in range(_target.shape[-1]):
            _selected = self._select_onestep(_current, min_candidate[..., i])
            optimum.append(_selected)
            _current = _selected

        result = u.Quantity(optimum).reshape(target_shape)
        return self._validate(result, _target)  # type: ignore

    def _select_onestep(
        self, current: u.Quantity, min_candidate: u.Quantity
    ) -> u.Quantity:
        """Select the optimum from candidates, one-step in time series."""
        if current.shape != min_candidate.shape:
            raise ValueError(
                "`current` and `min_candidate` must have same shape, got"
                f"'{current.shape}' and '{min_candidate.shape}'"
            )

        def __select(mc: u.Quantity) -> u.Quantity:
            nonlocal current
            candidates = u.Quantity(
                list(utils.frange(mc, self.limit.upper, 360 << u.deg))  # type: ignore
            )
            current = self._select(current, candidates)
            return current

        optimum = u.Quantity([__select(mc) for mc in min_candidate.flat])
        return optimum.reshape(min_candidate.shape)  # type: ignore

    def _select(self, current: u.Quantity, candidates: u.Quantity) -> u.Quantity:
        if candidates.ndim != 1:
            raise ValueError(
                f"Candidates should be 1-d array, got array of shape {current.shape}"
            )
        if candidates.size == 0:
            return float("nan") << u.deg  # type: ignore
        if candidates.size == 1:
            return candidates.reshape(())  # type: ignore

        distance = abs(candidates - current)
        sorted_indices_by_distance = distance.argsort()
        for idx in sorted_indices_by_distance:
            if distance[idx] < self.max_observation_size:
                return candidates[idx]  # type: ignore
            if candidates[idx] in self.preferred_limit:
                return candidates[idx]  # type: ignore
        return float("nan") << u.deg  # type: ignore

    def _validate(self, result: u.Quantity, target: u.Quantity) -> Optional[u.Quantity]:
        is_not_finite = ~np.isfinite(result)
        if is_not_finite.any():
            self.logger.warning(
                f"{is_not_finite.sum()} of instructed coordinates are out of drive "
                f"range {self.limit} : {target[is_not_finite]}"
            )
            return

        if self.preferred_limit.contain_all(result):
            return result
        self.logger.warning("Command position nears drive range limit.")
        return result



__all__ = ["PathFinder", "standby_position"]

import math
import time as pytime
from typing import Callable, Generator, List, Optional, Tuple, TypeVar, Union

import astropy.units as u
import numpy as np
from astropy.coordinates import BaseCoordinateFrame

from .. import config, utils
from ..typing import Number, UnitType
from .convert import CoordCalculator

T = TypeVar("T", Number, u.Quantity)


def standby_position(
    start: Tuple[T, T],
    end: Tuple[T, T],
    *,
    margin: Optional[Union[float, int, u.Quantity]],
    unit: Optional[UnitType] = None,
) -> Tuple[u.Quantity, u.Quantity]:
    """Calculate the standby position taking into account the margin during scanning.

    Parameters
    ----------
    start
        Longitude and latitude of start point.
    end
        Longitude and latitude of end point.
    unit
        Angular unit in which longitude and latitude are given. If they are given as
        ``Quantity``, this parameter will be ignored.

    Examples
    --------
    >>> neclib.coordinates.standby_position(start=(30, 50), end=(32, 50), unit="deg")
    (<Quantity 29. deg>, <Quantity 50. deg>)

    """
    margin = utils.get_quantity(margin, unit=unit)
    start_lon, start_lat = utils.get_quantity(start[0], start[1], unit=unit)
    end_lon, end_lat = utils.get_quantity(end[0], end[1], unit=unit)

    if start_lon != end_lon and start_lat == end_lat:  # scan along longitude
        standby_lon = start_lon - margin * np.sign(end_lon - start_lon)
        standby_lat = start_lat
    elif start_lon == end_lon and start_lat != end_lat:  # scan along latitude
        standby_lon = start_lon
        standby_lat = start_lat - margin * np.sign(end_lat - start_lat)
    else:
        raise NotImplementedError(
            "Diagonal scan isn't implemented yet."
        )  # TODO: Implement.

    return (standby_lon, standby_lat)


def get_position_angle(start_lon, end_lon, start_lat, end_lat):
    if end_lon == start_lon:
        position_angle = (np.pi / 2 * u.rad) * np.sign(end_lat - start_lat)
    else:
        position_angle = np.arctan((end_lat - start_lat) / (end_lon - start_lon))
        position_angle += (np.pi if end_lon < start_lon else 0) * u.rad

    return position_angle


class Timer:
    def __init__(self):
        self.start = pytime.time()
        self.target = self.start

    def set_offset(self, /, offset) -> None:
        if not isinstance(offset, (int, float)):
            raise TypeError(f"Offset must be int or float, not {type(offset)}")
        self.target += offset

    def get(self) -> float:
        return self.target

    def __bool__(self) -> bool:
        return True


class PathFinder(CoordCalculator):
    """望遠鏡の軌道を計算する

    Parameters
    ----------
    location
        Location of observatory.
    pointing_param_path
        Path to pointing parameter file.
    pressure
        Atmospheric pressure at the observation environment.
    temperature
        Temperature at the observation environment.
    relative_humidity
        Relative humidity at the observation environment.
    obswl
        Observing wavelength.

    Attributes
    ----------
    location: EarthLocation
        Location of observatory.
    pressure: Quantity
        Atmospheric pressure, to compute diffraction correction.
    temperature: Quantity
        Temperature, to compute diffraction correction.
    relative_humidity: Quantity or float
        Relative humidity, to compute diffraction correction.
    obswl: Quantity
        Observing wavelength, to compute diffraction correction.
    obsfreq: Quantity
        Observing frequency of EM-wave, to compute diffraction correction.

    Examples
    --------
    >>> location = EarthLocation("138.472153deg", "35.940874deg", "1386m")
    >>> path = "path/to/pointing_param.toml"
    >>> pressure = 850 * u.hPa
    >>> temperature = 290 * u.K
    >>> humid = 0.5
    >>> obsfreq = 230 * u.GHz
    >>> finder = neclib.coordinates.PathFinder(
    ...     location, path, pressure=pressure, temperature=temperature,
    ...     relative_humidity=humid, obsfreq=obsfreq)

    """

    command_unit_duration_sec = 1

    @property
    def unit_n_cmd(self) -> int:
        """Number of commands to be calculated in ``self.command_unit_duration_sec``."""
        return int(self.command_unit_duration_sec * config.antenna_command_frequency)

    def functional(
        self,
        lon: Callable[[float], T],
        lat: Callable[[float], T],
        frame: Union[str, BaseCoordinateFrame],
        *,
        unit: Optional[UnitType] = None,
        n_cmd: Union[int, float],
        time: Optional[Timer],
    ) -> Generator[Tuple[u.Quantity, u.Quantity, List[float]], None, None]:
        """Calculate the path of antenna movement from any function.

        Parameters
        ----------
        lon
            Function that returns longitude. The function should accept single argument;
            auxiliary variable of value between 0 and 1.
        lat
            Function that returns latitude.
        frame
            Frame in which the coordinates are given.
        unit
            Angular unit in which longitude and latitude are given. If they are given as
            ``Quantity``, this parameter will be ignored.
        n_cmd
            Number of commands. The path will be calculated supplying arithmetic
            sequence of auxiliary variable, which first term is 0, last is 1, number of
            members is ``n_cmd``.
        time
            Start time manager. If ``None``, current time will be used. No consideration
            for ``antenna_command_offset_sec`` is required.

        Returns
        -------
        Az, El, Time
            Tuple of conversion result, azimuth, elevation and time.

        Examples
        --------
        >>> def lon(x):
        ...     return 30 + x * 2
        >>> def lat(x):
        ...     return 50
        >>> path = finder.functional(lon, lat, frame="altaz", unit="deg", n_cmd=1000)
        >>> next(path)
        (<Quantity [30., 30.1, 30.2, ...] deg>, <Quantity [50., 50., 50., ...] deg>,
        [1610612736.0, 1610612736.1, 1610612736.2, ...])

        """
        if time is None:
            start = pytime.time() + config.antenna_command_offset_sec
        else:
            start = time.get() + config.antenna_command_offset_sec

        for seq in range(math.ceil(n_cmd / self.unit_n_cmd)):
            idx = [seq * self.unit_n_cmd + j for j in range(self.unit_n_cmd)]
            param = [_idx / n_cmd for _idx in idx]
            param = list(filter(lambda x: x <= 1, param))
            idx = idx[: len(param)]
            if (seq == math.ceil(n_cmd / self.unit_n_cmd) - 1) and (param[-1] < 1):
                param.append(1)
                idx.append(idx[-1] + (param[-1] - param[-2]) * n_cmd)

            lon_for_this_seq = [lon(p) for p in param]
            lat_for_this_seq = [lat(p) for p in param]
            t_for_this_seq = [
                start + (i / config.antenna_command_frequency) for i in idx
            ]

            yield self.get_altaz(
                lon=lon_for_this_seq,
                lat=lat_for_this_seq,
                frame=frame,
                unit=unit,
                obstime=t_for_this_seq,
            )

    def linear(
        self,
        start: Tuple[T, T],
        end: Tuple[T, T],
        frame: Union[str, BaseCoordinateFrame],
        *,
        speed: Union[float, int, u.Quantity],
        unit: Optional[UnitType] = None,
        time: Optional[Timer] = None,
    ) -> Generator[Tuple[u.Quantity, u.Quantity, List[float]], None, None]:
        """望遠鏡の直線軌道を計算する

        Parameters
        ----------
        start
            Longitude and latitude of start point.
        end
            Longitude and latitude of end point.
        frame
            Coordinate frame, in which longitude and latitude are given.
        speed
            Telescope drive speed.
        unit
            Angular unit in which longitude and latitude are given. If they are given as
            ``Quantity``, this parameter will be ignored.

        Returns
        -------
        Az, El, Time
            Tuple of calculated azimuth, elevation and time commands.

        Examples
        --------
        >>> finder.linear(
            start=(0, 0), end=(0.05, 0), frame="altaz", speed=0.5, unit=u.deg
        )
        [<Quantity [-1.47920569, -1.46920569, -1.45920569, -1.44920569, -1.43920569,
           -1.42920569] deg>, <Quantity [-1.88176239, -1.88176188, -1.88176136,
           -1.88176084, -1.88176032, -1.8817598 ] deg>, array([1.66685637e+09,
           1.66685637e+09, 1.66685637e+09, 1.66685637e+09,
           1.66685637e+09, 1.66685637e+09])]

        """
        start = utils.get_quantity(*start, unit=unit)
        end = utils.get_quantity(*end, unit=unit)
        speed = utils.get_quantity(speed, unit=end[0].unit / u.s)
        distance = ((start[0] - end[0]) ** 2 + (start[1] - end[1]) ** 2) ** 0.5
        n_cmd = (distance / speed) * (config.antenna_command_frequency * u.Hz)

        def lon(x):
            return start[0] + x * (end[0] - start[0])

        def lat(x):
            return start[1] + x * (end[1] - start[1])

        return self.functional(
            lon, lat, frame, unit=unit, n_cmd=float(n_cmd), time=time
        )

    def linear_with_acceleration(
        self,
        start: Tuple[T, T],
        end: Tuple[T, T],
        frame: Union[str, BaseCoordinateFrame],
        *,
        speed: Union[float, int, u.Quantity],
        unit: Optional[UnitType] = None,
        margin: Union[float, int, u.Quantity],
        time: Optional[Timer] = None,
    ) -> Generator[Tuple[u.Quantity, u.Quantity, List[float]], None, None]:
        margin_start = standby_position(start, end, margin=margin, unit=unit)
        margin_end = utils.get_quantity(*start, unit=unit)
        margin = utils.get_quantity(margin, unit=unit)
        speed = utils.get_quantity(speed, unit=margin.unit / u.s)

        position_angle = get_position_angle(
            margin_start[0], margin_end[0], margin_start[1], margin_end[1]
        )

        # マージン部分の座標計算 加速度その1
        a = (speed**2) / (2 * margin)

        # マージン部分の座標計算 加速度その2
        # a_az = config.antenna_max_acceleration_az
        # a_el = config.antenna_max_acceleration_el
        # a = (a_az**2 + a_el**2) ** (1 / 2)

        required_time = ((2 * margin) / a) ** (1 / 2)
        n_cmd = required_time.to_value("s") * config.antenna_command_frequency

        def lon(x):
            t = x * required_time
            a_lon = a * np.cos(position_angle)
            return margin_start[0] + a_lon * t**2 / 2

        def lat(x):
            t = x * required_time
            a_lat = a * np.sin(position_angle)
            return margin_start[1] + a_lat * t**2 / 2

        time = time or Timer()
        yield from self.functional(lon, lat, frame, unit=unit, n_cmd=n_cmd, time=time)
        time.set_offset(required_time.to_value("s"))
        yield from self.linear(start, end, frame, speed=speed, unit=unit, time=time)

    def offset_linear(
        self,
    ) -> Generator[Tuple[u.Quantity, u.Quantity, List[float]], None, None]:
        ...
        raise NotImplementedError

    def offset_track(
        self,
    ) -> Generator[Tuple[u.Quantity, u.Quantity, List[float]], None, None]:
        ...
        raise NotImplementedError

    def track(
        self,
        lon: T,
        lat: T,
        frame: Union[str, BaseCoordinateFrame],
        *,
        unit: Optional[UnitType] = None,
        time: Optional[Timer] = None,
    ) -> Generator[Tuple[u.Quantity, u.Quantity, List[float]], None, None]:
        time = time or Timer()
        while True:
            yield from self.functional(
                lambda _: lon,
                lambda _: lat,
                frame,
                unit=unit,
                n_cmd=self.unit_n_cmd,
                time=time,
            )
            time.set_offset(self.command_unit_duration_sec)

    def track_by_name(
        self, name: str, *, time: Optional[Timer] = None
    ) -> Generator[Tuple[u.Quantity, u.Quantity, List[float]], None, None]:
        ...
        raise NotImplementedError



import logging
import os
from typing import Dict, Optional

from ..utils import clip


class ColorizeLevelNameFormatter(logging.Formatter):
    """Colorize severity level name.

    See `logging docs <https://docs.python.org/3/library/logging.html#handler-objects>`_
    for the usage.

    """

    ColorPrefix: Dict[int, str] = {
        0: "\x1b[0m",  # NOTSET, default (Black or White)
        10: "\x1b[35m",  # DEBUG, Magenta
        20: "\x1b[32m",  # INFO, Green
        30: "\x1b[33m",  # WARNING, Yellow
        40: "\x1b[31m",  # ERROR, Red
        50: "\x1b[41;97m",  # CRITICAL, White on Red
    }

    def format(self, record: logging.LogRecord) -> str:
        """Format a record to text."""
        levelno = int(clip(record.levelno // 10 * 10, 0, 50))  # type: ignore
        original_levelname = record.levelname
        record.levelname = self.ColorPrefix[levelno] + original_levelname + "\x1b[0m"
        return super().format(record)


class ConsoleLogger(logging.Logger):

    OBSERVE_level = logging.INFO + 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        logging.addLevelName(self.OBSERVE_level, "OBSERVE")

    def obslog(self, msg, indent_level: int = 0, *args, **kwargs):
        """Log observation summary.

        Argument other than ``indent_level`` are interpreted as for
        `logging.Logger.debug <https://docs.python.org/3/library/logging.html#logging.Logger.debug>`_

        indent_level
            If non-zero, this message is logged with indentation.

        """  # noqa: E501
        indented_msg = "    " * indent_level + msg
        super()._log(self.OBSERVE_level, indented_msg, args, **kwargs)


def get_logger(
    name: Optional[str] = None,
    min_level: int = None,
) -> ConsoleLogger:
    """Get logger instance which prints operation logs to console.

    Parameters
    ----------
    name
        Name of the logger. Calling this function with same ``name`` returns the same
        logger instance.
    min_level
        Lower bound of severity level to be displayed on terminal. To suppress less
        severe messages, set higher value. No matter this value, the log file contains
        all messages severer than ``logging.DEBUG`` (level=10).

    Examples
    --------
    >>> logger = neclib.interfaces.getLogger("OTF_observation", "path/to/log.txt")
    >>> logger.debug("Inform something only needed on diagnosing some problem.")
    >>> logger.info("Inform something EXPECTED has happened.")
    >>> logger.warning("Inform something user should care has happened.")
    >>> logger.error(f"Notify some {functionality} cannot be completed due to error.")
    >>> logger.critical(
    ...     "Notify the %s cannot continue operation due to %s", "program", "some error"
    ... )
    >>> logger.obslog("Observation finished", indent_level=1)

    """
    logging.setLoggerClass(ConsoleLogger)
    logger_name = "neclib" if name is None else f"neclib.{name.strip('neclib.')}"
    logger = logging.getLogger("necst." + logger_name)

    min_level = (
        int(os.environ.get("NECST_LOG_LEVEL", logging.INFO))
        if min_level is None
        else min_level
    )

    fmt = "%(asctime)-s: [%(levelname)-s: %(filename)s#L%(lineno)s] %(message)s"
    color_log_format = ColorizeLevelNameFormatter(fmt)

    ch = logging.StreamHandler()
    ch.setLevel(min_level)
    ch.setFormatter(color_log_format)

    rootLogger = logging.getLogger()
    chs = [_ch for _ch in rootLogger.handlers if isinstance(_ch, logging.StreamHandler)]
    [rootLogger.handlers.remove(_ch) for _ch in chs]
    rootLogger.addHandler(ch)

    return logger  # type: ignore



import os
import time
from datetime import datetime
from pathlib import Path
from threading import Event, Thread
from typing import List, Optional, Union

from .. import get_logger
from .writer_base import Writer


class Recorder:
    """Data recorder, to be used with arbitrary writers.

    Parameters
    ----------
    record_root
        Root directory of data storage. All the data will be stored inside this
        directory. Subdirectories will be structured by attached writers.

    Examples
    --------
    >>> recorder = neclib.recorders.Recorder("/home/user/data")
    >>> recorder.add_writer(neclib.recorders.DBWriter(), neclib.recorders.FileWriter())
    >>> recorder.start_recording()
    >>> recorder.append("test", 1)  # Arbitrary number and type of data can be passed
    >>> recorder.stop_recording()

    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, record_root: Path) -> None:
        self.__writers: List[Writer] = []
        self.record_root = Path(record_root)
        self.recording_path = None

        self.logger = get_logger(self.__class__.__name__)

        self._thread: Optional[Thread] = None
        self._event: Optional[Event] = None

    def add_writer(self, *writers: Writer) -> None:
        """Attach writer(s) to this recorder."""
        if any(type(w) is type for w in writers):
            raise TypeError("Writer should be instantiated.")
        self.__writers.extend(writers)

    @property
    def writers(self):
        """List of attached writers."""
        return self.__writers

    def drop_writer(self, *writers: Union[int, Writer]) -> None:
        """Drop writer(s) from this recorder."""
        for writer in writers:
            if isinstance(writer, int):
                to_remove = self.__writers.pop(writer)
                to_remove.stop_recording()
            else:
                writer.stop_recording()
                self.__writers.remove(writer)

    def start_recording(
        self, record_dir: Optional[os.PathLike] = None, *, noreset: bool = False
    ) -> None:
        """Activate all attached writers."""
        if self.is_recording:
            return
        if record_dir is not None:
            if (self._thread is not None) or (self._event is not None):
                raise RuntimeError(
                    "Cannot start named recording with background check running. "
                    "Please stop the recorder first."
                )
            self.recording_path = self.record_root / Path(record_dir)
        else:
            self.recording_path = self._auto_generate_record_dir()

            if not noreset:
                self._thread = Thread(target=self._check_db_date, daemon=True)
                self._event = Event()
                self._thread.start()

        for writer in self.__writers:
            writer.start_recording(self.recording_path)

    def append(self, *args, **kwargs) -> None:
        """Pass data to all attached writers."""
        if self.recording_path is None:
            raise RuntimeError("Recorder not started. Incoming data won't be kept.")
        handled = [writer.append(*args, **kwargs) for writer in self.__writers]
        if not any(handled):
            err_msg = f"No writer handled the data: {args, kwargs}"
            self.logger.warning(err_msg[slice(0, min(100, len(err_msg)))])

    def stop_recording(self, *, noreset: bool = False) -> None:
        """Deactivate all attached writers."""
        if not self.is_recording:
            return

        if not noreset:
            if self._event is not None:
                self._event.set()
            if self._thread is not None:
                self._thread.join()
            self._thread = self._event = None

        for writer in self.__writers:
            writer.stop_recording()
        self.recording_path = None

    def _auto_generate_record_dir(self) -> Path:
        now = datetime.utcnow()
        record_dir = self.record_root / now.strftime("%Y%m") / now.strftime("%Y%m%d")
        record_dir.mkdir(parents=True, exist_ok=True)
        return record_dir

    def _check_db_date(self) -> None:
        if self._event is None:
            return
        while not self._event.is_set():
            if self.recording_path != self._auto_generate_record_dir():
                self.stop_recording(noreset=True)
                self.start_recording(noreset=True)
            time.sleep(1)

    @property
    def is_recording(self) -> bool:
        """Whether this recorder is accepting data or not."""
        return self.recording_path is not None



__all__ = ["skip_on_simulator"]

import functools


def skip_on_simulator(func):
    from .. import config  # Avoid circular import

    @functools.wraps(func)
    def _skip_if_simulator(*args, **kwargs):
        if config.simulator:
            return None
        return func(*args, **kwargs)

    return _skip_if_simulator



from typing import List

import astropy.units as u

from ...utils import busy
from .bias_reader_base import BiasReader


class CPZ3177(BiasReader):
    Manufacturer = "Interface"
    Model = "CPZ3177"

    Identifier = "rsw_id"

    def __init__(self) -> None:
        import pyinterface

        self.rsw_id = self.Config.rsw_id
        self.ave_num = self.Config.ave_num
        self.smpl_freq = self.Config.smpl_freq
        self.single_diff = self.Config.single_diff
        self.all_ch_num = self.Config.all_ch_num
        self.smpl_ch_req = self.Config.smpl_ch_req
        self.ch_selector = self.Config.ch_selector
        self.v_mag = self.Config.v_mag
        self.i_mag = self.Config.i_mag

        self.ad = pyinterface.open(3177, self.rsw_id)
        self.ad.stop_sampling()
        self.ad.initialize()
        self.ad.set_sampling_config(
            smpl_ch_req=self.smpl_ch_req,  # It must be a list
            smpl_num=1000,
            smpl_freq=self.smpl_freq,
            single_diff=self.single_diff,
            trig_mode="ETERNITY",
        )
        self.ad.start_sampling("ASYNC")

    def get_data(self, ch) -> List[float]:
        with busy(self, "busy"):
            offset = self.ad.get_status()["smpl_count"] - self.ave_num
            data = self.ad.read_sampling_buffer(self.ave_num, offset)
            data_li_2 = []
            for i in range(self.all_ch_num):
                data_li = []
                for k in range(self.ave_num):
                    data_li.append(data[k][i])
                data_li_2.append(data_li)

            ave_data_li = []
            for data in data_li_2:
                d = sum(data) / self.ave_num
                ave_data_li.append(d)
            return ave_data_li[ch - 1]

    def get_bias_voltage(self, ch) -> u.Quantity:
        ch_name = "bias_ch{}_v".format(ch)
        pci_ch = self.ch_selector[ch_name]
        return self.get_data(pci_ch) * self.v_mag * u.mV

    # This "ch" which is argument of this function is ch of bias box, not pci board.

    def get_bias_current(self, ch) -> u.Quantity:
        ch_name = "bias_ch{}_i".format(ch)
        pci_ch = self.ch_selector[ch_name]
        return self.get_data(pci_ch) * self.i_mag * u.microampere

    # This "ch" which is argument of this function is ch of bias box, not pci board.

    def finalize(self) -> None:
        self.ad.stop_sampling()



import os
import queue
import struct
import sys
import time
import traceback
from contextlib import contextmanager
from threading import Event, Thread
from typing import Dict, Generator, List, Tuple

import xfftspy

from ... import get_logger
from .spectrometer_base import Spectrometer


class XFFTS(Spectrometer):

    Manufacturer: str = "Radiometer Physics GmbH"
    Model: str = "XFFTS"

    Identifier = "host"

    def __init__(self) -> None:
        self.logger = get_logger(self.__class__.__name__)

        self.host = self.Config.host
        self.cmd_port = self.Config.cmd_port
        self.data_port = self.Config.data_port
        self.synctime_us = self.Config.synctime_us
        self.bw_mhz = {int(k): v for k, v in self.Config.bw_mhz.items()}
        self.data_input, self.setting_output = self.initialize()

        self.data_queue = queue.Queue(maxsize=2)
        self.thread = None
        self.event = None
        self.start()

        self.warn = False

    def start(self) -> None:
        if (self.thread is not None) or (self.event is not None):
            self.stop()
        self.thread = Thread(target=self._read_data, daemon=True)
        self.event = Event()
        self.thread.start()

    def _read_data(self) -> None:
        while (self.event is not None) and (not self.event.is_set()):
            if self.data_queue.full():
                if self.warn:
                    self.logger.warning(
                        "Dropping the data due to low readout frequency."
                    )
                    self.warn = False
                self.data_queue.get()

            try:
                data = self.data_input.receive_once()
                self.data_queue.put((time.time(), data["data"]))
            except struct.error:
                exc = traceback.format_exc()
                self.logger.warning(exc[slice(0, min(len(exc), 100))])

    def stop(self) -> None:
        if self.event is not None:
            self.event.set()
        if self.thread is not None:
            self.thread.join()
        self.event = self.thread = None
        self.warn = False

    def initialize(self) -> Tuple[xfftspy.data_consumer, xfftspy.udp_client]:
        """Get configured data input and setting output."""
        setting_output = xfftspy.udp_client(self.host, self.cmd_port, print=False)
        setting_output.stop()
        setting_output.set_synctime(self.synctime_us)  # synctime in us
        _sections = [int(i in self.bw_mhz) for i in range(1, max(self.bw_mhz) + 1)]
        setting_output.set_usedsections(_sections)
        for board_id, bw_mhz in self.bw_mhz.items():
            setting_output.set_board_bandwidth(board_id, bw_mhz)
        setting_output.configure()  # Apply settings
        setting_output.caladc()  # Calibrate ADCs
        setting_output.start()

        data_input = xfftspy.data_consumer(self.host, self.data_port)
        data_input.clear_buffer()
        return data_input, setting_output

    def get_spectra(self) -> Tuple[float, Dict[int, List[float]]]:
        self.warn = True
        return self.data_queue.get()

    def finalize(self) -> None:
        self.setting_output.stop()
        self.stop()


@contextmanager
def mute_stderr() -> Generator[None, None, None]:
    stderr = sys.stderr
    sys.stderr = open(os.devnull, "w")
    try:
        yield
    except Exception:
        pass
    finally:
        sys.stderr = stderr



r"""Pointing error correction.

The telescope pointing is critically important in performing accurate and uniform
observation. This module employs following pointing model:

.. math::

    \Delta x =& \chi_{Az} \sin ( \omega_{Az} - Az ) \sin ( El ) \\
    &+ \epsilon \sin ( El ) \\
    &+ \chi_{2, Az} \sin ( 2 ( \omega_{2, Az} - Az ) ) \sin ( El ) \\
    &+ \mathrm{d}Az \cos ( El ) \\
    &+ \mathrm{d}e \\
    &+ \mathrm{cor}_v \cos ( El + \mathrm{cor}_p ) \\
    &+ \mathrm{d}e_\mathrm{radio} \\
    \Delta Az =& \Delta x / \cos ( El ) \\
    \Delta y =& - \chi_{El} \cos ( \omega_{El} - Az ) \\
    &- \chi_{2, El} \cos ( 2 ( \omega_{2, El} - Az ) ) \\
    &+ g_1 \cos ( El ) + g_2 \sin ( El ) \\
    &+ \mathrm{d}el \\
    &+ g_{ 1,\mathrm{radio} } \cos ( El ) + g_{ 2,\mathrm{radio} } \sin ( El ) \\
    &- \mathrm{cor}_v \sin ( El + \mathrm{cor}_p ) \\
    &+ \mathrm{d}el_\mathrm{radio} \\
    \Delta El =& \Delta y

"""

__all__ = ["PointingError"]

from typing import Iterable, Optional, Tuple, Union

import astropy.units as u
import numpy as np

from .parser import PointingErrorData
from ..typing import UnitType


class PointingError(PointingErrorData):
    """Observation pointing error parameter calculator.

    Examples
    --------
    >>> params = neclib.parameters.PointingError.from_file("path/to/params.toml")
    >>> params.refracted2apparent(az=[10, 20, 30], el=[20, 30, 40], unit="deg")
    (<Quantity [11.3, 21.3, 31.3] deg>, <Quantity [20.5, 30.5, 40.5] deg>)
    >>> params.apparent2refracted(
    ...     az=Quantity([10, 20, 30], unit="deg"), el=Quantity([20, 30, 40], unit="deg")
    ... )
    (<Quantity [11.3, 21.3, 31.3] deg>, <Quantity [20.5, 30.5, 40.5] deg>)
    >>> params.apparent2refracted(
    ...     az=Quantity([10, 20, 30], unit="deg"),
    ...     el=Quantity([20, 30, 40], unit="deg"),
    ...     unit="arcsec"
    ... )
    (<Quantity [40680., 76680., 112680.] arcsec>, <Quantity [73800., 109800., 145800.] arcsec>)

    """  # noqa: E501

    def _get_offset(
        self, az: u.Quantity, el: u.Quantity
    ) -> Tuple[u.Quantity, u.Quantity]:
        """Error from calculated coordinates to encoder reading."""
        gravitational_term = np.polynomial.Polynomial(
            [0, self.g, self.gg, self.ggg, self.gggg]
        )
        radio_gravitational_term = np.polynomial.Polynomial(
            [0, self.g_radio, self.gg_radio, self.ggg_radio, self.gggg_radio]
        )

        dx = (
            self.chi_Az * np.sin(self.omega_Az - az) * np.sin(el)
            + self.eps * np.sin(el)
            + self.chi2_Az * np.sin(2 * (self.omega2_Az - az)) * np.sin(el)
            + self.dAz * np.cos(el)
            + self.de
            + self.cor_v * np.cos(el + self.cor_p)
            + self.de_radio
        )  # NOTE: 3rd term: chi2 sin sin, revised from chi2 cos sin
        dAz = dx / np.cos(el)
        dEl = (
            -1 * self.chi_El * np.cos(self.omega_El - az)
            - self.chi2_El * np.cos(2 * (self.omega2_El - az))
            + gravitational_term(el.to("deg").value) * u.arcsec
            + self.dEl
            + radio_gravitational_term(el.to("deg").value) * u.arcsec
            - self.cor_v * np.sin(el + self.cor_p)
            + self.del_radio
        )  # NOTE: 2nd term: chi2 cos, revised from chi2 sin
        # TODO: Review dimension of gravitational terms
        return dAz, dEl

    def _force_data_to_be_quantity(
        self,
        data: Union[u.Quantity, float, Iterable[float]],
        unit: Optional[UnitType] = None,
    ) -> u.Quantity:
        if isinstance(data, u.Quantity):
            return data if unit is None else data.to(unit)
        return u.Quantity(data, unit=unit)

    def refracted2apparent(
        self,
        az: Union[u.Quantity, float, Iterable[float]],
        el: Union[u.Quantity, float, Iterable[float]],
        *,
        unit: Optional[UnitType] = None,
    ) -> Tuple[u.Quantity, u.Quantity]:
        az = self._force_data_to_be_quantity(az, unit)
        el = self._force_data_to_be_quantity(el, unit)
        dAz, dEl = self._get_offset(az, el)
        return az - dAz, el - dEl  # type: ignore

    def apparent2refracted(
        self,
        az: Union[u.Quantity, float, Iterable[float]],
        el: Union[u.Quantity, float, Iterable[float]],
        *,
        unit: Optional[UnitType] = None,
    ) -> Tuple[u.Quantity, u.Quantity]:
        az = self._force_data_to_be_quantity(az, unit)
        el = self._force_data_to_be_quantity(el, unit)
        dAz, dEl = self._get_offset(az, el)
        return az + dAz, el + dEl  # type: ignore



# Core recorder
from .recorder import Recorder  # noqa: F401

# Writers
from .console_log_writer import ConsoleLogWriter  # noqa: F401
from .file_writer import FileWriter  # noqa: F401
from .necstdb_writer import NECSTDBWriter  # noqa: F401



import time
from pathlib import Path
from typing import Optional

from .. import get_logger
from ..typing import TextLike
from .writer_base import Writer


class FileWriter(Writer):
    def __init__(self) -> None:
        if not self._initialized[self.__class__]:
            self.logger = get_logger(self.__class__.__name__)
            self.record_dir: Optional[Path] = None
            self._initialized[self.__class__] = True

    def start_recording(self, record_dir: Path) -> None:
        self.record_dir = record_dir

    def append(self, path: str, contents: TextLike, *args, **kwargs) -> bool:
        if not (isinstance(path, str) and isinstance(contents, TextLike)):
            return False
        if self.record_dir is None:
            self.logger.warning("FileWriter is not recording")
            return False

        header = f"# Original file: {path}\n# Recorded time: {time.time()}\n"

        write_path = self._find_new_path(path)
        write_path.parent.mkdir(parents=True, exist_ok=True)
        if isinstance(contents, bytes):
            write_path.write_bytes(header.encode("utf-8") + contents)
        else:
            write_path.write_text(header + contents)  # type: ignore

        return True

    def stop_recording(self) -> None:
        self.record_dir = None

    def _find_new_path(self, path: str, i: int = 0) -> Path:
        if self.record_dir is None:
            raise RuntimeError("FileWriter is not recording")

        _path = Path(path.split(":")[-1])
        _path = _path if i == 0 else _path.parent / f"{_path.stem}({i}){_path.suffix}"
        _path: Path = self.record_dir / _path.name
        return self._find_new_path(path, i + 1) if _path.exists() else _path



import time

import astropy.units as u
import ogameasure

from ...units import dBm
from ...utils import busy, skip_on_simulator
from .signal_generator_base import SignalGenerator


class FSW0020(SignalGenerator):

    Manufacturer: str = "PhaseMatrix"
    Model = "FSW0020"

    Identifier = "host"

    @skip_on_simulator
    def __init__(self):
        com = ogameasure.ethernet(self.Config.host, self.Config.port)
        self.sg = ogameasure.Phasematrix.FSW0020(com)
        self.sg.use_external_reference_source()

    def set_freq(self, freq_GHz):
        with busy(self, "busy"):
            self.sg.freq_set(freq_GHz)
            time.sleep(1)
            return

    def set_power(self, power_dBm):
        """Set the power of the signal generator output.

        Attention
        ---------
        The ability to change power of this signal generator is optional configuration.
        For products without the option, this function has no effect but raises no
        error.

        """
        with busy(self, "busy"):
            self.sg.power_set(power_dBm)
            time.sleep(1)
            return

    def get_freq(self):
        with busy(self, "busy"):
            f = self.sg.freq_query()
            time.sleep(1)
            return f * u.Hz

    def get_power(self):
        """Get the power of the signal generator output.

        Attention
        ---------
        The ability to change power of this signal generator is optional configuration.
        For products without the option, this function may return meaningless value.

        """
        with busy(self, "busy"):
            f = self.sg.power_query()
            time.sleep(1)
            return f * dBm

    def start_output(self):
        with busy(self, "busy"):
            self.sg.output_on()
            time.sleep(1)
            return

    def stop_output(self):
        with busy(self, "busy"):
            self.sg.output_off()
            time.sleep(1)
            return

    def get_output_status(self):
        with busy(self, "busy"):
            f = self.sg.output_query()
            time.sleep(1)
            if f == 1:
                return True
            elif f == 0:
                return False
            else:
                return None

    def finalize(self):
        self.stop_output()
        try:
            self.sg.com.close()
        except AttributeError:
            pass



import importlib
import sys
from pathlib import Path

from . import selector

paths = Path(__file__).parent.iterdir()
module_paths = filter(lambda p: p.is_dir() and p.name[0] not in "._", paths)
impl_modules = [
    importlib.import_module(f".{m.name}", __package__) for m in module_paths
]

implementations = selector.list_implementations(impl_modules)
"""List of all available implementations."""

parsed = selector.parse_device_configuration(impl_modules)
"""List of parsed device implementations."""

here = sys.modules[__name__]
_ = [setattr(here, k, v) for k, v in parsed.items()]



__all__ = ["describe_frame", "parse_frame"]

import re
from dataclasses import dataclass
from functools import lru_cache
from typing import Type, Union

from astropy.coordinates import (
    Angle,
    BaseCoordinateFrame,
    SkyOffsetFrame,
    frame_transform_graph,
)


@dataclass
class Frame:

    frame: Union[str, BaseCoordinateFrame, Type[BaseCoordinateFrame]]

    def __post_init__(self):
        if isinstance(self.frame, str):
            self.frame = self._parse(self.frame)

    def _parse(
        self, frame: str
    ) -> Union[BaseCoordinateFrame, Type[BaseCoordinateFrame]]:
        _parsed = frame_transform_graph.lookup_name(frame.lower())
        if _parsed is not None:
            return _parsed

        _parsed = re.match(
            r"origin\s*=\s*([a-z0-9]*)\(([a-z\d\s.]*),\s*([a-z\d\s.]*)\),\s*"
            r"rotation\s*=\s*([a-z\d\s.]*)",
            frame.lower(),
        )
        if _parsed is not None:
            base_frame, lon, lat, rotation = _parsed.groups()
            BaseFrame = frame_transform_graph.lookup_name(base_frame.lower())
            rotation = Angle(rotation)
            if BaseFrame is None:
                raise ValueError(f"Unknown frame {base_frame!r}")
            return SkyOffsetFrame(origin=BaseFrame(lon, lat), rotation=rotation)
        raise ValueError(f"Could not parse frame {frame!r}")

    def _describe(
        self, frame: Union[BaseCoordinateFrame, Type[BaseCoordinateFrame]]
    ) -> str:
        if "offset" not in frame.name:
            return frame.name
        center = frame.origin
        attr_names = center.representation_component_names
        attr_name_mapping = {v: k for k, v in attr_names.items()}
        lon = getattr(center, attr_name_mapping["lon"])
        lat = getattr(center, attr_name_mapping["lat"])
        return f"origin={center.name}({lon}, {lat}), rotation={frame.rotation}"

    @classmethod
    @lru_cache(maxsize=16)
    def from_string(cls, frame: str) -> "Frame":
        return cls(frame)

    def __str__(self) -> str:
        if "offset" not in self.frame.name:
            return self.frame.name
        return self._describe(self.frame)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._describe(self.frame)})"


def parse_frame(frame: str) -> Union[SkyOffsetFrame, Type[BaseCoordinateFrame]]:
    """Parse a frame string into a frame object.

    Parameters
    ----------
    frame
        The frame string to parse.

    Returns
    -------
    frame
        The frame object.

    Examples
    --------
    >>> neclib.coordinates.parse_frame("fk5")
    <class 'astropy.coordinates.builtin_frames.fk5.FK5'>
    >>> neclib.coordinates.parse_frame("altaz")
    <class 'astropy.coordinates.builtin_frames.altaz.AltAz'>
    >>> neclib.coordinates.parse_frame("origin=FK5(10deg, 0deg), rotation=20deg")
    <SkyOffsetFK5 Frame (equinox=J2000.000, rotation=20.0 deg, origin=<FK5 Coordinate
    (equinox=J2000.000): (ra, dec) in deg
    (10., 0.)>)>

    """
    return Frame.from_string(frame).frame


def describe_frame(frame: Union[BaseCoordinateFrame, Type[BaseCoordinateFrame]]) -> str:
    """String representation of a frame.

    Parameters
    ----------
    frame
        The frame to describe.

    Returns
    -------
    repr
        String representation of the frame.

    Examples
    --------
    >>> neclib.coordinates.describe_frame(FK5)
    'fk5'
    >>> neclib.coordinates.describe_frame(AltAz())
    'altaz'
    >>> neclib.coordinates.describe_frame(
    ...     SkyOffsetFrame(origin=FK5("10deg", "0deg"), rotation="20deg")
    ... )
    'origin=fk5(10deg, 0deg), rotation=20deg'

    """
    return str(Frame(frame))



"""Convert flexibly specified observation parameters to some useful form.

.. note::

   Public API won't change, but contents such as ``translate_coordsys``,
   ``COORDINATES``, and offset calculation in ``off_point_coord`` can be moved to other
   subpackage or module, i.e. ``neclib.coordinates``?

"""

__all__ = ["ObsParams", "interval", "off_point_coord"]

from typing import Any, ClassVar, Dict, Hashable, List, Optional, Tuple, Union

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord
from astropy.coordinates.baseframe import BaseCoordinateFrame

from .parser import ObsParamData
from .. import units as custom_u
from ..typing import CoordFrameType, Number, UnitType
from ..utils import ParameterMapping, quantity2builtin

# Only those not supported by Astropy
COORDINATES = {
    "j2000": "fk5",
    "b1950": "fk4",
    "horizontal": "altaz",
}


def translate_coordsys(coordsys: str) -> str:
    return COORDINATES.get(coordsys.lower(), coordsys.lower())


class ObsParams(ObsParamData):
    r"""Observation parameter calculator.

    Parameters
    ----------
    units
        Parameter name to its unit mapping.
    **kwargs
        Parameters.

    Attributes
    ----------
    val: ParameterMapping
        Returns built-in type value if the unit is provided in ``units``.

    Examples
    --------
    >>> neclib.parameters.ObsParams.ParameterUnit = {
    ...     "deg": ["LamdaOn", "BetaOn", ...], "s": ...
    ... }
    >>> params = neclib.parameters.ObsParams.from_file("path/to/parameter.file")
    >>> params.LamdaOn
    <Angle 123.45 deg>
    >>> params.val.LamdaOn
    123.45

    """

    ParameterName: Dict[str, Union[str, List[str]]] = {
        "off_observation_interval": "off_interval",
        "hot_observation_interval": "load_interval",
        "on_point_coordinate": ["LamdaOn", "BetaOn", "COORD_SYS"],
        "offset_from_on_to_off": ["deltaLamda", "deltaBeta", "DELTA_COORD"],
        "offset_coslat_applied": "OTADEL",
        "absolute_off_point_coordinate": ["LamdaOff", "BetaOff", "COORD_SYS"],
    }
    ParameterUnit: ClassVar[
        Union[Dict[Hashable, UnitType], Dict[UnitType, List[str]]]
    ] = {}

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.val = ParameterMapping(
            quantity2builtin(self, self.ParameterUnit)  # type: ignore
        )

    def _getitem(self, param_type: str) -> Union[Any, List[Any]]:
        keys = self.ParameterName[param_type]

        if isinstance(keys, str):
            return self.get(keys, None)
        params = [self.get(key, None) for key in keys]
        if not any([param is None for param in params]):
            return params

    def off_observation_interval(
        self, *, unit: Union[str, u.Unit], points_per_scan: Optional[int] = None
    ) -> Tuple[Number, str]:
        """Parse OFF observation interval.

        See Also
        --------
        interval : Implementation of this method.

        """
        return interval(
            self._getitem("off_observation_interval"),  # type: ignore
            unit=unit,
            points_per_scan=points_per_scan,
        )

    def hot_observation_interval(
        self, *, unit: Union[str, u.Unit], points_per_scan: Optional[int] = None
    ) -> Tuple[Number, str]:
        """Parse HOT observation interval.

        See Also
        --------
        interval : Implementation of this method.

        """
        return interval(
            self._getitem("hot_observation_interval"),  # type: ignore
            unit=unit,
            points_per_scan=points_per_scan,
        )

    def off_point_coord(
        self, *, unit: Union[str, u.Unit], **kwargs
    ) -> Tuple[Number, Number, str]:
        """Parse HOT observation interval.

        See Also
        --------
        off_point_coord : Implementation of this method.

        """

        def translate(coordinate):
            if coordinate is None:
                return
            lon, lat, coordsys = coordinate
            return lon, lat, translate_coordsys(coordsys)

        return off_point_coord(
            unit=unit,
            on_point=translate(self._getitem("on_point_coordinate")),
            offset=translate(self._getitem("offset_from_on_to_off")),
            coslat_applied=self._getitem("offset_coslat_applied"),  # type: ignore
            absolute=translate(self._getitem("absolute_off_point_coordinate")),
            **kwargs,
        )


def interval(
    quantity: u.Quantity,
    *,
    unit: Union[str, u.Unit],
    points_per_scan: Optional[int] = None,
) -> Tuple[Number, str]:
    """Parse interval.

    Parameters
    ----------
    quantity
        Input value to be parsed.
    unit
        Unit in which return value is given.
    points_per_scan
        Number of observation points per 1 scan line.

    Examples
    --------
    >>> neclib.parameters.interval(Quantity("5min"), unit="s")
    300, "time"
    >>> neclib.parameters.interval(Quantity("5scan"), unit="point", points_per_scan=5)
    25, "point"

    """
    if quantity.unit.is_equivalent("second"):  # type: ignore
        return quantity.to_value(unit), "time"
    elif quantity.unit is u.Unit(unit):  # type: ignore
        return quantity.value, str(unit)

    if points_per_scan is None:
        return quantity.to_value(unit), str(unit)  # type: ignore
    converted = quantity.to_value(
        unit, equivalencies=custom_u.scan_to_point_equivalency(points_per_scan)
    )
    return converted, str(unit)


def off_point_coord(
    *,
    unit: Union[str, u.Unit],
    on_point: Optional[Tuple[u.Quantity, u.Quantity, CoordFrameType]] = None,
    offset: Optional[Tuple[u.Quantity, u.Quantity, CoordFrameType]] = None,
    coslat_applied: Optional[bool] = None,
    absolute: Optional[Tuple[u.Quantity, u.Quantity, CoordFrameType]] = None,
    **kwargs,
) -> Tuple[Number, Number, str]:
    """Parse OFF observation coordinate.

    Give (``absolute`` coordinate of OFF point, output ``unit``) or (coordinate of
    ``on_point``, ``offset`` from the ON point, if the offset value has
    ``coslat_applied``, output ``unit``). The coordinates should be given in the
    form: (longitude, latitude, coordinate frame)

    Parameters
    ----------
    on_point
        Coordinate of ON point.
    offset
        Offset from ON point to OFF point. Caution the unit, as ``Angle("0h0m1s")``
        is not equivalent to ``Angle("1arcsec")``.
    coslat_applied
        If ``False``, longitude of ``offset`` will be divided by cos(latitude).
    absolute
        Absolute coordinate of OFF point.
    unit
        Unit in which return value is given.
    **kwargs
        Keyword arguments to ``SkyCoord``, will be passed to coordinate frame class
        such as AltAz; location, obstime, obswl, etc.

    Examples
    --------
    >>> neclib.parameters.off_point_coord(
    ...     absolute=(Angle("3h25m20s"), Angle("31d32m33s"), "fk5"),
    ...     unit="deg",
    ... )
    (51.3333, 31.5425, 'fk5')
    >>> neclib.parameters.off_point_coord(
    ...     on_point=(Angle("5h1m1s"), Angle("30d25m20s"), "galactic"),
    ...     offset=(Angle("0 0 10 hours"), Angle("20arcsec"), "fk5"),
    ...     coslat_applied=True,
    ...     unit="deg",
    ... )
    (266.6442444454571, 48.39027455272732, 'fk5')

    """
    # Indices for tuple of length 3.
    LON, LAT, COORDSYS = 0, 1, 2

    def _convert_unit(
        x: u.Quantity, y: u.Quantity, coordsys: Union[str, BaseCoordinateFrame]
    ) -> Tuple[Number, Number, str]:
        # Extract name of ``astropy.coordinates.baseframe.BaseCoordinateFrame``.
        # The reason ``getattr`` is used instead of ``isinstance`` or ``issubclass``
        # is either of the class or its instance can be given and handling them is
        # quite complicated.
        coordsys = getattr(coordsys, "name", coordsys)
        return x.to_value(unit), y.to_value(unit), coordsys.lower()

    if (absolute is not None) and (offset is None):
        return _convert_unit(*absolute)

    if (absolute is None) and (offset is not None) and (on_point is not None):
        on_coord = (
            SkyCoord(on_point[LON], on_point[LAT], frame=on_point[COORDSYS], **kwargs)
            .transform_to(offset[COORDSYS].lower())
            .data
        )
        if not isinstance(coslat_applied, bool):
            raise TypeError("Explicitly give ``coslat_applied`` or not.")
        if not coslat_applied:
            offset = (
                offset[LON] / np.cos(on_coord.lat),  # type: ignore
                offset[LAT],
                offset[COORDSYS],
            )
        return _convert_unit(
            on_coord.lon + offset[LON],  # type: ignore
            on_coord.lat + offset[LAT],  # type: ignore
            offset[COORDSYS],
        )

    raise ValueError(
        "Coordinate to calculate is ambiguous."
        "Give either of ``absolute`` or ``offset``, not both."
    )



import logging
from pathlib import Path
from typing import Optional

from ..interfaces.console_logger import ConsoleLogger
from .writer_base import Writer


class ConsoleLogWriter(Writer):
    """Record writer for console log.

    Attributes
    ----------
    log_file_path
        Path to file into which all logs (severity >= ``logging.DEBUG``) are dumped.
    obslog_file_path
        Path to file to log observation summary.

    """

    def __init__(self) -> None:
        if not self._initialized[self.__class__]:
            fmt = "%(asctime)-s: [%(levelname)-s: %(filename)s#L%(lineno)s] %(message)s"
            self.log_format = logging.Formatter(fmt)
            self.obslog_format = logging.Formatter("- [(UTC) %(asctime)-s] %(message)s")

            self.log_file_path: Optional[Path] = None
            self.obslog_file_path: Optional[Path] = None

            self._initialized[self.__class__] = True

    def start_recording(self, record_dir: Path) -> None:
        record_dir.mkdir(parents=True, exist_ok=True)

        self.log_file_path = record_dir / "console.log"
        self.obslog_file_path = record_dir / "observation.log"

        self.fh = logging.FileHandler(self.log_file_path)
        self.fh.setLevel(logging.DEBUG)
        self.fh.setFormatter(self.log_format)

        self.obs_fh = logging.FileHandler(self.obslog_file_path)
        self.obs_fh.addFilter(
            lambda record: record.levelno == ConsoleLogger.OBSERVE_level
        )
        self.obs_fh.setFormatter(self.obslog_format)

        rootLogger = logging.getLogger()
        rootLogger.addHandler(self.fh)
        rootLogger.addHandler(self.obs_fh)

    def append(self, *args, **kwargs) -> bool:
        """This writer don't accept any data not issued by ``logging.Logger``."""
        return False

    def stop_recording(self) -> None:
        self.log_file_path = None
        self.obslog_file_path = None

        rootLogger = logging.getLogger()
        rootLogger.removeHandler(self.fh)
        rootLogger.removeHandler(self.obs_fh)



"""Type aliases for simple type hinting."""

__all__ = ["PathLike", "AngleUnit"]

import os
from typing import Callable, Literal, Protocol, Tuple, Union, runtime_checkable

import astropy.units as u
import numpy as np
from astropy.coordinates import BaseCoordinateFrame


PathLike = Union[os.PathLike, str]
"""Alias of ``os.PathLike``, with ``str`` type combined."""

AngleUnit = Literal["deg", "arcmin", "arcsec", "rad"]
"""Literal expression of supported angular units in this package."""

Number = Union[int, float, np.ndarray]
"""Dimensionless value types."""
ValueType = Number
"""Dimensionless value types."""

Boolean = Union[bool, np.bool_]
"""Boolean types."""
BoolType = Boolean
"""Boolean types."""

Unit = Union[str, u.UnitBase]
"""Unit of physical quantity."""
UnitType = Unit
"""Unit of physical quantity."""

CoordFrameType = Union[str, BaseCoordinateFrame]
"""Coordinate frame type."""

QuantityValue = Union[Number, u.Quantity]
"""Physical quantity or primitive type with unit separately specified."""

EquivalencyType = Tuple[
    u.UnitBase, u.UnitBase, Callable[[float], float], Callable[[float], float]
]
"""Type expression for unit equivalency in astropy.units."""


@runtime_checkable
class TextLike(Protocol):
    def upper(self) -> "TextLike":
        ...

    def lower(self) -> "TextLike":
        ...

    def find(self) -> int:
        ...

    def replace(self) -> "TextLike":
        ...

    def __len__(self) -> int:
        ...


@runtime_checkable
class SupportsComparison(Protocol):
    def __eq__(self, other: "SupportsComparison") -> bool:
        ...

    def __ne__(self, other: "SupportsComparison") -> bool:
        ...

    def __lt__(self, other: "SupportsComparison") -> bool:
        ...

    def __le__(self, other: "SupportsComparison") -> bool:
        ...

    def __gt__(self, other: "SupportsComparison") -> bool:
        ...

    def __ge__(self, other: "SupportsComparison") -> bool:
        ...
